#ifndef _RXERROR_H_
#define _RXERROR_H_

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <iostream.h>
#include <RxDlist.h>
#include <RxCio.h>
#include <thrdatadefs.h>

#ifndef NO_EXCEPTIONS
#define THROW(x)	throw (x);
#else
#define THROW(x)
#endif

class RxError : public RxDnodeBase
{
	friend class RxErrCtx;
    public:
	enum action {
	   DEFAULT,
	   IGNORE, 
	   THROW,
	   ABORT,
	   EXIT,
	   POST };
	struct CodeList {
	   int code;
	   char *str;
	};
    private:
	// used internally to assign an integer index to a new
	// class derived from RxError
	static int assignClassIdx(int *idxp);

	static int idx;		// integer index for this class
	action action_;		// action: what to do when error raised

    protected:
	int code;		// error code
	char *msg;		// error message

	void initAction();	// called inside each constructor
	virtual void dothrow()	// throw an exception of this type
	 { THROW(*this)
	 };
	virtual RxError *docopy()
	 { return new RxError(*this);
	 };
	virtual const CodeList *getCodeList() const
	 { return NULL;
	 }
	static int getClassIdx(int *idxp);
	int getClassIdx() const; // get integer index for this class
	virtual int *getClassIdxPtr() const
	 { return &idx;
	 }	
	static action setActionIdx (action act, int *idxp);
	static action getActionIdx (int *idxp);

    public:
	RxError();
	RxError(int code, const char *fmt, ...);
	RxError(int code);
	RxError(const RxError &rec);
	~RxError();
	void vset (int n, const char *fmt, va_list ap);
	void set (int n, const char *fmt, ...);
	void setMsg (const char *s);
		
	const char *getMsg() const
	 { return msg;
	 };
	int setCode(int c)
	 { int old = code; code = c; return old;
	 }
	int getCode() const
	 { return code;
	 };
	void raise();
	void post();

	const char *codeToStr(int code) const;
	virtual const char *getName() const
	 { return "RxError";
	 }
	static action setAction(action act)
	 { return setActionIdx (act, &idx);
	 }
	static action getAction()
	 { return getActionIdx (&idx);
	 }
	action netAction() const
	 { return action_;
	 }
	RX_DNODE_DECLS (RxError);
};

typedef enum RxError::action RxErrorAction;

class RxErrCtx
{
    public:
	friend class RxError;
    private:
	static RxErrCtx *ctx0;	// used in the single thread instance
	static THRDATA_KEY key;	// used in the multi-thread instance
	static int keyinit;

	RxDlist<RxError> elist;
#define DEFAULT_NUM_CLASSES	16
	RxErrorAction *classActions;
	int numClasses;

	RxError *err;
	
	RxErrorAction setClassAction (int idx, RxErrorAction action);
	RxErrorAction getClassAction (int idx);
    public:
	static RxErrCtx *getCur();

	int print (FILE *fp)
	 { RxCoutput co(fp);
	   return print(&co);
	 }
	int print (ostream *osp)
	 { RxCoutput co(osp);
	   return print(&co);
	 }
	int print (char *buf, int n=0)
	 { RxCoutput co(buf, n);
	   return print(&co);
	 }
	int print (RxCoutput *co);
	int numErrs ();
	int clear ();

	RxError *firstErr()
	 { return err = elist.head();
	 }
	RxError *nextErr()
	 { return (err ? err = err->next() : NULL);
	 }
	RxError *lastErr()
	 { return err = elist.tail();
	 }
	RxError *prevErr()
	 { return (err ? err = err->prev() : NULL);
	 }

	static int printCur(FILE *fp)
	 { return getCur()->print(fp);
	 }
	static int printCur (ostream *osp)
	 { return getCur()->print(osp);
	 }
	static int printCur (char *buf, int n=0)
	 { return getCur()->print(buf, n);
	 }
	static int numErrsCur()
	 { return getCur()->numErrs();
	 }
	static int clearCur()
	 { return getCur()->clear();
	 }

	static RxError *firstErrCur()
	 { return getCur()->firstErr();
	 }
	static RxError *nextErrCur()
	 { return getCur()->nextErr();
	 }
	static RxError *lastErrCur()
	 { return getCur()->lastErr();
	 }
	static RxError *prevErrCur()
	 { return getCur()->prevErr();
	 }

	RxErrCtx ();
	~RxErrCtx ();
};

#define RX_ERROR_DECLS(className) \
    private: \
	static CodeList codelist[]; \
	static int idx; \
    protected: \
	virtual void dothrow() \
	 { THROW(*this) \
	 }; \
	virtual RxError *docopy() \
	 { return new className(*this); \
	 }; \
	virtual const CodeList *getCodeList() const \
	 { return codelist; \
	 }; \
	virtual int *getClassIdxPtr() const \
	 { return &idx; \
	 } \
    public: \
	virtual const char *getName() const \
	 { return # className; \
	 }; \
	static action setAction (action act) \
	 { return setActionIdx (act, &idx); \
	 } \
	static action getAction () \
	 { return getActionIdx (&idx); \
	 }

#define RX_ERROR_CONSTRUCTOR_CODE() \
	initAction();

#define RX_ERROR_DEFS_BEGIN(T) \
int T::idx = 0; \
RxError::CodeList T::codelist[] = {

#define RX_ERROR_DEFS_END() \
	0,          NULL };
#endif
