#include <RxQtrn.h>
#include <RxRot.h>
#include <RxMathError.h>
#include <fastmath.h>
#include <math.h>
#include <stdarg.h>

#define EPS	(1e-13)

const RxQtrn RxQtrn::ZERO(0,0,0,0);
const RxQtrn RxQtrn::IDENT(1,0,0,0);

double& RxQtrn::operator()(int i)
{
	static double dum;

	switch (i)
	 { case 0:
	    { return c;
	    }
	   case 1:
	    { return v.x();
	    }
	   case 2:
	    { return v.y();
	    }
	   case 3:
	    { return v.z();
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxQtrn::(%d): %K\n", i).raise();
	      return (dum);
	    }
	 }
}

double RxQtrn::operator()(int i) const
{
	static double dum;

	switch (i)
	 { case 0:
	    { return c;
	    }
	   case 1:
	    { return v.x();
	    }
	   case 2:
	    { return v.y();
	    }
	   case 3:
	    { return v.z();
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxQtrn::(%d): %K\n", i).raise();
	      return (dum);
	    }
	 }
}

RxQtrn& RxQtrn::setRot (const RxRot &R1)
{
	/*
	   Based on the follwoing algorithm:

	   Let the elements of the matrix be given by Q_ij. For i,j,k
	   such that epsilon (i,j,k) = 1, use the maximum of 

	   q_i = 1/2 sqrt (1 + Q_ii - Q_jj - Q_kk)
	   q   = 1/2 sqrt (1 + Q_ii + Q_jj + Q_kk)

	   and the relations

	   q_j q_k = 1/4 (Q_kj + Q_jk)
	   q q_i   = 1/4 (Q_kj - Q_jk)

	   We use this algorithm because it is numerically well behaved.
	*/

#define C_MAX	1
#define X_MAX   2
#define Y_MAX   3
#define Z_MAX   4

	int max;
	double sumC, sumX, sumY, sumZ;
	double s;

	sumC = R1.n.x() + R1.o.y() + R1.a.z();
	sumX = R1.n.x() - R1.o.y() - R1.a.z();
	sumY = R1.o.y() - R1.a.z() - R1.n.x();
	sumZ = R1.a.z() - R1.n.x() - R1.o.y();

	if (sumC > sumX)
	 { if (sumY > sumZ)
	    { max = (sumC > sumY) ? C_MAX : Y_MAX;
	    }
	   else
	    { max = (sumC > sumZ) ? C_MAX : Z_MAX;
	    }
	 }
	else
	 { if (sumY > sumZ)
	    { max = (sumX > sumY) ? X_MAX : Y_MAX;
	    }
	   else
	    { max = (sumX > sumZ) ? X_MAX : Z_MAX;
	    }
	 }

	switch (max)
	 { case C_MAX:
/*		printf ("C_MAX\n");	*/

		s = sqrt (1.0 + sumC) * 0.5;
		c = s;
	        s *= 4.0;
		v.x() = (R1.o.z() - R1.a.y()) / s;
		v.y() = (R1.a.x() - R1.n.z()) / s;
		v.z() = (R1.n.y() - R1.o.x()) / s;
	        break;
	   case X_MAX:
/*		printf ("X_MAX\n");	*/

		s = sqrt (1.0 + sumX) * 0.5;
		v.x() = s;
		s *= 4.0;
		c =   (R1.o.z() - R1.a.y()) / s;
	        v.y() = (R1.o.x() + R1.n.y()) / s;
		v.z() = (R1.a.x() + R1.n.z()) / s;
		break;
	   case Y_MAX:
/*		printf ("Y_MAX\n");	*/

		s = sqrt (1.0 + sumY) * 0.5;
		v.y() = s;
		s *= 4.0;
		c =   (R1.a.x() - R1.n.z()) / s;
	        v.x() = (R1.o.x() + R1.n.y()) / s;
		v.z() = (R1.o.z() + R1.a.y()) / s;
		break;
	   case Z_MAX:
/*		printf ("Z_MAX\n");	*/

		s = sqrt (1.0 + sumZ) * 0.5;
		v.z() = s;
		s *= 4.0;
		c =   (R1.n.y() - R1.o.x()) / s;
	        v.y() = (R1.o.z() + R1.a.y()) / s;
		v.x() = (R1.a.x() + R1.n.z()) / s;
		break;
	 }
	if (c < 0.0)
	 { c = -c;
	   v.x() = -v.x();
	   v.y() = -v.y();
	   v.z() = -v.z();
	 }
	return (*this);
}

void RxQtrn::getRot (RxRot *R1) const
{
	/* This is based on a algorithm in Ken Shoemake's paper

		``Animating Rotation with Quaternion Curves''
		SIGGRAPH 1985.

	   It sets the rotational component of t to be equal to the
	   rotation described by the quaternion q.
	*/

	double twoX, twoY, twoZ;
	double twoXW, twoXX, twoXY;
	double twoYW, twoYY, twoYZ;
	double twoZW, twoZZ, twoZX;

	twoX = v.x() + v.x();
	twoY = v.y() + v.y();
	twoZ = v.z() + v.z();

	twoXW = twoX * c;
	twoYW = twoY * c;
	twoZW = twoZ * c;
	twoXX = twoX * v.x();
	twoYY = twoY * v.y();
	twoZZ = twoZ * v.z();
	twoXY = twoX * v.y();
	twoYZ = twoY * v.z();
	twoZX = twoZ * v.x();

	R1->n.x() = 1.0 - twoYY - twoZZ;
	R1->n.y() = twoXY + twoZW;
	R1->n.z() = twoZX - twoYW;
	R1->o.x() = twoXY - twoZW;
	R1->o.y() = 1.0 - twoXX - twoZZ;
	R1->o.z() = twoYZ + twoXW;
	R1->a.x() = twoZX + twoYW;
	R1->a.y() = twoYZ - twoXW;
	R1->a.z() = 1.0 - twoXX - twoYY;
}

RxQtrn &RxQtrn::setEul (double phi, double the, double psi)
{
	RxRot R;

	return (setRot (R.setEul (phi, the, psi)));
}

void RxQtrn::getEul (double *phip, double *thep, double *psip) const
{
	RxRot R;

	getRot (&R);
	R.getEul (phip, thep, psip);
}

RxQtrn &RxQtrn::setRpy (double roll, double pitch, double yaw)
{
	RxRot R;

	return (setRot (R.setRpy (roll, pitch, yaw)));
}

void RxQtrn::getRpy (double *rollp, double *pitchp, double *yawp) const
{
	RxRot R;

	getRot (&R);
	R.getRpy (rollp, pitchp, yawp);
}

RxQtrn& RxQtrn::setAxis (const RxVecd3 &vec, double ang)
{
	double s;

	c = cos(ang/2);
	s = sin(ang/2);
	v.unit (vec);
	v.scale (s);
	return (*this);
}

void RxQtrn::getAxis (RxVecd3 *vec, double *angp) const
{
	// normalize, just in case

	double mag;
	double vmag;

	vmag = sqrt(v.dotQ(v));
	mag = sqrt(c*c + vmag*vmag);
	
	mag = sqrt(c*c + v.dotQ(v));
	vec->scaleQ (1/vmag, v);
	if (angp) *angp = 2*atan2 (vmag/mag, c/mag);	
}

RxQtrn &RxQtrn::setVecs (const char *axes,const RxVecd3 &v1,const RxVecd3 &v2)
{
	RxRot R;

	return (setRot (R.setVecs (axes, v1, v2)));
}

RxQtrn &RxQtrn::setAngs (const char *axes, ...)
{
	RxRot R;

	va_list ap;
	va_start (ap, axes);
	R.vsetAngs (axes, /*inDegrees=*/0, ap);
	va_end (ap);
	return (setRot (R));
}

RxQtrn &RxQtrn::setAngsD (const char *axes, ...)
{
	RxRot R;

	va_list ap;
	va_start (ap, axes);
	R.vsetAngs (axes, /*inDegrees=*/1, ap);
	va_end (ap);
	return (setRot (R));
}

RxQtrn& RxQtrn::add (const RxQtrn &q1, const RxQtrn &q2)
{
	c = q1.c + q2.c;
	v.addQ (q1.v, q2.v);
	return (*this);
}

RxQtrn& RxQtrn::sub (const RxQtrn &q1, const RxQtrn &q2)
{
	c = q1.c - q2.c;
	v.subQ (q1.v, q2.v);
	return (*this);
}

RxQtrn& RxQtrn::mul (const RxQtrn &q1, const RxQtrn &q2)
{
	RxVecd3 vtmp;
	double ctmp;

	ctmp = q1.c*q2.c - q1.v.dotQ(q2.v);
	vtmp.cross (q1.v, q2.v);
	vtmp.extrapQ (vtmp, q1.c, q2.v);
	v.extrapQ (vtmp, q2.c, q1.v);
	c = ctmp;
	return (*this);
}

RxQtrn& RxQtrn::mulI (const RxQtrn &q1, const RxQtrn &q2)
{
	RxVecd3 vtmp;
	double ctmp;
	double d;

	d = q2.c*q2.c + q2.v.dotQ(q2.v);

	ctmp = q1.c*q2.c + q1.v.dotQ(q2.v);
	vtmp.cross (q2.v, q1.v);
	vtmp.extrapQ (vtmp, -q1.c, q2.v);
	vtmp.extrapQ (vtmp, q2.c, q1.v);
	v.scale (1/d, vtmp);
	c = ctmp;
	return (*this);
}

RxQtrn& RxQtrn::Imul (const RxQtrn &q1, const RxQtrn &q2)
{
	RxVecd3 vtmp;
	double ctmp;
	double d;
	
	d = q1.c*q1.c + q1.v.dotQ(q1.v);

	ctmp = q1.c*q2.c + q1.v.dotQ(q2.v);
	vtmp.cross (q2.v, q1.v);
	vtmp.extrapQ (vtmp, q1.c, q2.v);
	vtmp.extrapQ (vtmp, -q2.c, q1.v);
	v.scale (1/d, vtmp);
	c = ctmp;
	return (*this);
}

RxQtrn& RxQtrn::ImulI (const RxQtrn &q1, const RxQtrn &q2)
{
	RxVecd3 vtmp;
	double ctmp;
	double d;

	d = (q1.c*q1.c + q1.v.dotQ(q1.v))*(q2.c*q2.c + q2.v.dotQ(q2.v));
	
	ctmp = q1.c*q2.c - q1.v.dotQ(q2.v);
	vtmp.cross (q1.v, q2.v);
	vtmp.extrapQ (vtmp, -q1.c, q2.v);
	vtmp.extrapQ (vtmp, -q2.c, q1.v);
	v.scale (1/d, vtmp);
	c = ctmp;
	return (*this);
}

#if 1
RxQtrn& RxQtrn::slerp (const RxQtrn &q1, const RxQtrn &q2, double u)
{
	RxQtrn qp;
	double ang;
	double s0, c0;
	double su, cu;
	double a2;

	c0 = q1.c*q2.c + q1.v.dotQ(q2.v);
	qp.c = q2.c - c0*q1.c;
	qp.v.extrapQ (q2.v, -c0, q1.v);
	s0 = sqrt(qp.c*qp.c + qp.v.dotQ(qp.v));

	ang = ATAN2 (s0, c0);
	SINCOS(&su, &cu, ang*u);
	if (s0 < EPS)
	 { a2 = 0;
	   cu = 1;
	 }
	else
	 { a2 = su/s0;
	 }
	c = cu*q1.c + a2*qp.c;
	v.x() = cu*q1.v.x() + a2*qp.v.x();
	v.y() = cu*q1.v.y() + a2*qp.v.y();
	v.z() = cu*q1.v.z() + a2*qp.v.z();
	return (*this);
}
#endif

#if 0
RxQtrn& RxQtrn::slerp (const RxQtrn &q1, const RxQtrn &q2, double u)
{
	RxQtrn qd;
	double ang;
	double s;

	qd.Imul (q1, q2);
	s = qd.v.norm2();
	if (s > EPS)
	 { ang = ATAN2 (s, qd.c);
	   qd.c = COS(ang*u);
	   qd.v.scale (SIN(ang*u)/s);
	   mul (q1, qd);
	 }
	else
	 { qd.sub (q2, q1);
	   qd.scale (u);
	   add (q1, qd);
	   unit ();
	 }
	return (*this);
}
#endif

RxQtrn& RxQtrn::mulC (const RxQtrn &q1, const RxQtrn &q2)
{
	RxVecd3 vtmp;
	double ctmp;

	ctmp = q1.c*q2.c + q1.v.dotQ(q2.v);
	vtmp.cross (q2.v, q1.v);
	vtmp.extrapQ (vtmp, -q1.c, q2.v);
	v.extrapQ (vtmp, q2.c, q1.v);
	c = ctmp;
	return (*this);
}

RxQtrn& RxQtrn::Cmul (const RxQtrn &q1, const RxQtrn &q2)
{
	RxVecd3 vtmp;
	double ctmp;

	ctmp = q1.c*q2.c + q1.v.dotQ(q2.v);
	vtmp.cross (q2.v, q1.v);
	vtmp.extrapQ (vtmp, q1.c, q2.v);
	v.extrapQ (vtmp, -q2.c, q1.v);
	c = ctmp;
	return (*this);
}

RxQtrn& RxQtrn::CmulC (const RxQtrn &q1, const RxQtrn &q2)
{
	RxVecd3 vtmp;
	double ctmp;

	ctmp = q1.c*q2.c - q1.v.dotQ(q2.v);
	vtmp.cross (q1.v, q2.v);
	vtmp.extrapQ (vtmp, -q1.c, q2.v);
	v.extrapQ (vtmp, -q2.c, q1.v);
	c = ctmp;
	return (*this);
}

RxQtrn& RxQtrn::scale (double a, const RxQtrn &q1)
{
	c = q1.c * a;
	v.scaleQ (a, q1.v);
	return (*this);
}

RxQtrn& RxQtrn::unit(const RxQtrn &q1)
{
	double mag;

	mag = sqrt(q1.c*q1.c + q1.v.dotQ(q1.v));
	c = q1.c/mag;
	v.scaleQ (1/mag, q1.v);
	return (*this);
}

RxQtrn& RxQtrn::inv (const RxQtrn &q1)
{
	double d;

	d = q1.c*q1.c + q1.v.dotQ(q1.v);
	c = q1.c/d;
	v.scaleQ (-1/d, q1.v);
	return (*this);
}

double RxQtrn::norm2 () const
{
	return (sqrt (c*c + v.dotQ(v)));
}

int RxQtrn::equal (const RxQtrn &Q1, double eps)
{
	RxQtrn Qneg;

	if (RxVecdBase::equal (Q1, eps))
	 { return 1;
	 }
	Qneg.neg (Q1);
	if (RxVecdBase::equal (Qneg, eps))
	 { return 1;
	 }
	else
	 { return 0;
	 }
}
