#include <RxRot.h>
#include <RxMathError.h>
#include <stdarg.h>
#include <math.h>
#include <cdefs.h>
#include <RxParseObj.h>
#include <RxSysError.h>

#define SMALL	(1e-12)

static double identBuf[9] = {
	1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 };
const RxRot RxRot::IDENT(identBuf);

void RxRot::doCopy(const RxRot &m1)
{
	n.x() = m1.n.x();
	n.y() = m1.n.y();
	n.z() = m1.n.z();
	o.x() = m1.o.x();
	o.y() = m1.o.y();
	o.z() = m1.o.z();
	a.x() = m1.a.x();
	a.y() = m1.a.y();
	a.z() = m1.a.z();
}

double& RxRot::operator()(int i, int j)
{
	static double dum;

	switch (j)
	 { case 0:
	    { return (n(i));
	    }
	   case 1:
	    { return (o(i));
	    }
	   case 2:
	    { return (a(i));
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxRot::(%d,%d): %K\n", i, j).raise();
	      return dum;
	    }
	 }
}

double RxRot::operator()(int i, int j) const
{
	static double dum;

	switch (j)
	 { case 0:
	    { return (n(i));
	    }
	   case 1:
	    { return (o(i));
	    }
	   case 2:
	    { return (a(i));
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
			   "RxRot(%d,%d): %K\n", i, j).raise();
	      return dum;
	    }
	 }
}

RxRot &RxRot::set (const double *dp)
{
	n.x() = dp[0];
	o.x() = dp[1];
	a.x() = dp[2];
	n.y() = dp[3];
	o.y() = dp[4];
	a.y() = dp[5];
	n.z() = dp[6];
	o.z() = dp[7];
	a.z() = dp[8];
	return (*this);
}

void RxRot::get (double *dp) const
{
	dp[0] = n.x();
	dp[1] = o.x();
	dp[2] = a.x();
	dp[3] = n.y();
	dp[4] = o.y();
	dp[5] = a.y();
	dp[6] = n.z();
	dp[7] = o.z();
	dp[8] = a.z();
}

RxRot &RxRot::setEul (double phi, double the, double psi)
{
	/* Set the rotation as specified by the input Euler angles. */
	   
	double sphi, sthe, spsi, cphi, cthe, cpsi;

	sphi = sin(phi);
	cphi = cos(phi);
	sthe = sin(the);
	cthe = cos(the);
	spsi = sin(psi);
	cpsi = cos(psi);

	n.x() =   cphi * cthe * cpsi - sphi * spsi;
	n.y() =   sphi * cthe * cpsi + cphi * spsi;
	n.z() = - sthe * cpsi;

	o.x() = - cphi * cthe * spsi - sphi * cpsi;
	o.y() = - sphi * cthe * spsi + cphi * cpsi;
	o.z() =   sthe * spsi;

	a.x() = cphi * sthe;
	a.y() = sphi * sthe;
	a.z() = cthe;
	return(*this);
}

RxRot &RxRot::setRpy (double roll, double pitch, double yaw)
{
	/* Set the rotational as specified by the roll-pitch-yaw angles. */
	   
	double sroll, spitch, syaw, croll, cpitch, cyaw;

	sroll = sin(roll);
	croll = cos(roll);
	spitch = sin(pitch);
	cpitch = cos(pitch);
	syaw = sin(yaw);
	cyaw = cos(yaw);

	n.x() = croll * cpitch;
	n.y() = sroll * cpitch;
	n.z() = - spitch;
	o.x() = croll * spitch * syaw - sroll * cyaw;
	o.y() = sroll * spitch * syaw + croll * cyaw;
	o.z() = cpitch * syaw;
	a.x() = croll * spitch * cyaw + sroll * syaw;
	a.y() = sroll * spitch * cyaw - croll * syaw;
	a.z() = cpitch * cyaw;
	return(*this);
}

RxRot &RxRot::setAxis (const RxVecd3 &v, double ang)
{
	/* set the rotation equal to a rotation 'h' about vector 'v'. */

	RxVecd3 kl;
	double s, c, ver, x, y, z, xv, yv, zv, xs, ys, zs;

       	kl.unit(v);
	s = sin(ang);
	ver = 1. - (c = cos(ang));
	xv = ver * (x = kl.x());
	xs = x * s;
	yv = ver * (y = kl.y());
	ys = y * s;
	zv = ver * (z = kl.z());
	zs = z * s;

	n.x() = x * xv + c;
	n.y() = x * yv + zs;
	n.z() = x * zv - ys;

	o.x() = y * xv - zs;
	o.y() = y * yv + c;
	o.z() = y * zv + xs;

	a.x() = z * xv + ys;
	a.y() = z * yv - xs;
	a.z() = z * zv + c;

	return(*this);
}

RxRot &RxRot::setVecs (const char *axes, const RxVecd3 &v1, const RxVecd3 &v2)
{
	/* Takes a pair of non-parallel axes
	   and uses them to build the rotation part of a TRSF structure.
	   v1 and v2 should describe two of the three final axes. Which
	   axis each vector describes is given by a code character ('x', 'y',
	   or 'z') in the two-character string 'axes'. 

	   v1 and v2 do not have to be orthogonal. v1 is assumed to be
	   parallel to the axis specified by axes[0]. The projection of
	   v2 onto the plane defined by this axis is then assumed to be
	   parallel to the axis specified by axes[1]. The third axis is
	   taken to be the cross product of the other two axes.

	   The routine raises RxMathError::BadSpec in the event of a 
	   specification error, or if the two defining vectors are very 
	   close to parallel.
	 */

	register RxVecd3 *r1, *r2;
	register double mag;

	RxVecd3 *r3;
	int negate;

	/* Determine which input vector corresponds to which axis */

	if (axes[0] == 'x')
	 { r1 = &n;
	   if (axes[1] == 'y')
	    { r2 = &o;
	      r3 = &a;
	      negate = 0;
	    }
	   else if (axes[1] == 'z')
	    { r2 = &a;
	      r3 = &o;
	      negate = 1;
	    }
	   else
	    { goto badString;
	    }
	 }
	else if (axes[0] == 'y')
	 { r1 = &o;
	   if (axes[1] == 'x')
	    { r2 = &n;
	      r3 = &a;
	      negate = 1;
	    }
	   else if (axes[1] == 'z')
	    { r2 = &a;
	      r3 = &n;
	      negate = 0;
	    }
	   else
	    { goto badString;
	    }
	 }
	else if (axes[0] == 'z')
	 { r1 = &a;
	   if (axes[1] == 'x')
	    { r2 = &n;
	      r3 = &o;
	      negate = 0;
	    }
	   else if (axes[1] == 'y')
	    { r2 = &o;
	      r3 = &n;
	      negate = 1;
	    }
	   else
	    { goto badString;
	    }
	 }
	else
	 { goto badString;
	 }

	if ((mag = sqrt(v1.x()*v1.x() + v1.y()*v1.y() + v1.z()*v1.z())) < SMALL)
	 { RxMathError (RxMathError::BadSpec,
	       "RxRot::setVecs(): %K v1 very small\n").raise();
	   return (*this);
	 }
	r1->x() = v1.x()/mag;
	r1->y() = v1.y()/mag;
	r1->z() = v1.z()/mag;
	mag = (v2.x()*r1->x() + v2.y()*r1->y() + v2.z()*r1->z());
	r2->x() = v2.x() - r1->x()*mag;
	r2->y() = v2.y() - r1->y()*mag;
	r2->z() = v2.z() - r1->z()*mag;
	if ((mag = sqrt(r2->x()*r2->x() + r2->y()*r2->y() + r2->z()*r2->z())) < SMALL)
	 { RxMathError (RxMathError::BadSpec,
	       "RxRot::setVecs(): %K v1Xv2 very small\n").raise();
	   return (*this);
	 }
	r2->x() = r2->x()/mag;
	r2->y() = r2->y()/mag;
	r2->z() = r2->z()/mag;
	if (negate)
	 { r3->x() = r1->z()*r2->y() - r1->y()*r2->z();
	   r3->y() = r1->x()*r2->z() - r1->z()*r2->x();
	   r3->z() = r1->y()*r2->x() - r1->x()*r2->y();
	 }
	else
	 { r3->x() = r1->y()*r2->z() - r1->z()*r2->y();
	   r3->y() = r1->z()*r2->x() - r1->x()*r2->z();
	   r3->z() = r1->x()*r2->y() - r1->y()*r2->x();
	 }
	return(*this);
badString:
	RxMathError (RxMathError::BadSpec,
		     "RxRot::setVecs(): %K %s\n", (const char*)axes).raise();
	return(*this);
}

int RxRot::vsetAngs (const char *axes, int inDegrees, va_list ap)
{
	double s, c, ang;
	const char *cp;
	RxRot rot;

	this->setIdent();

	for (cp=axes; *cp != '\0'; cp++)
	 { rot.setIdent();

	   ang = va_arg (ap, double);
	   if (inDegrees)
	    { ang *= DTOR;
	    }
	   c = cos(ang);
	   s = sin(ang);
	   switch (*cp)
	    { case 'x':
	       { rot.o.y() =  c;
		 rot.a.y() = -s;
		 rot.o.z() =  s;
		 rot.a.z() =  c;
		 break;
	       }
	      case 'y':
	       { rot.n.x() =  c;
		 rot.a.x() =  s;
		 rot.n.z() = -s;
		 rot.a.z() =  c;
		 break;
	       }
	      case 'z':
	       { rot.n.x() =  c;
		 rot.o.x() = -s;
		 rot.n.y() =  s;
		 rot.o.y() =  c;
		 break;
	       }
	      default:
	       { RxMathError (RxMathError::BadSpec,
		    "RxRot::setAngs %K %s\n", (const char*)axes).raise();
		 return (-1);
	       }
	    }
	   (*this) *= rot;
	 }
	return (1);
}

RxRot &RxRot::setAngs (const char *axes, ...)
{
	va_list ap;
	va_start (ap, axes);
	vsetAngs (axes, /*inDegrees=*/0, ap);
	va_end (ap);
	return (*this);
}

RxRot &RxRot::setAngsD (const char *axes, ...)
{
	va_list ap;
	va_start (ap, axes);
	vsetAngs (axes, /*inDegrees=*/1, ap);
	va_end (ap);
	return (*this);
}

void RxRot::getEul (double *phi, double *the, double *psi) const
{
	/* Return the Euler angles associated with the rotational
	   component of tr.
	 */
	double sphi, cphi, p;

	if (ABS(a.y()) > SMALL || ABS(a.x()) > SMALL) {
		*phi = (p = atan2(a.y(), a.x()));
		sphi = sin(p);
		cphi = cos(p);
		*the = 
		       atan2(cphi * a.x() + sphi * a.y(),
			     a.z());
		*psi = 
		       atan2(- sphi * n.x() + cphi * n.y(),
			     - sphi * o.x() + cphi * o.y());
	} 
	else {
		*phi = 0.;
		*the = atan2(a.x(), a.z());
		*psi = atan2(n.y(), o.y());
	}
}

void RxRot::getRpy (double *roll, double *pitch, double *yaw) const
{
	/* Return the roll-pitch-yaw angles associated with the rotational
	   component of tr.
	 */
	double sroll, croll, nx, ny, p;

	nx = n.x();
	ny = n.y();
	if (ABS(nx) < SMALL && ABS(ny) < SMALL) {
		*roll = 0.;
		*pitch = atan2(- n.z(), nx);
		*yaw = atan2(- a.y(), o.y());
	} 
	else {
		*roll = (p = atan2(ny, nx));
		sroll = sin(p);
		croll = cos(p);
		*pitch = 
			 atan2(- n.z(), croll * nx + sroll * ny);
		*yaw =   
			 atan2(sroll * a.x() - croll * a.y(),
			       croll * o.y() - sroll * o.x());
	}
}

#define X_MAX	1
#define Y_MAX	2
#define Z_MAX	3
#undef SMALL
#define SMALL (1e-14)

void RxRot::getAxis (RxVecd3 *v, double *ang) const
{
	/* Return the axis and angle of rotation associated with the 
	   rotational component of tr.
	 */

	register double cosine, tmpX, tmpY, tmpZ, tmpS;
	double versine, sine;
	int max;

	cosine = 0.5*(n.x() + o.y() + a.z() - 1.0);

	if (cosine > 0.0)
	 { /* then we can compute things the easy way. */

	   tmpX = o.z() - a.y();
	   tmpY = a.x() - n.z();
	   tmpZ = n.y() - o.x();

	   tmpS = sqrt (tmpX*tmpX + tmpY*tmpY + tmpZ*tmpZ);

	   if (tmpS < SMALL)
	    { *ang = v->y() = v->z() = 0.0;
	      v->x() = 1.0;
	      return;
	    }
	   v->x() = tmpX/tmpS;	      
	   v->y() = tmpY/tmpS;	      
	   v->z() = tmpZ/tmpS;
	   sine = tmpS / 2.0;
	 }
	else
	 { versine = 1.0 - cosine;

	   tmpX = ABS (n.x() - cosine);
	   tmpY = ABS (o.y() - cosine);
	   tmpZ = ABS (a.z() - cosine);
   
	   if (tmpX >= tmpY)
	    { max = (tmpX >= tmpZ ? X_MAX : Z_MAX);
	    }
	   else
	    { max = (tmpY >= tmpZ ? Y_MAX : Z_MAX);
	    }
	   switch (max)
	    { case X_MAX:
	       { if (versine < SMALL || tmpX <= 0.0)
		  { *ang = v->y() = v->z() = 0.0;
		    v->x() = 1.0;
		    return;
		  }
		 if ((tmpS = o.z() - a.y()) >= 0.0)
		  { v->x() = sqrt(tmpX/versine);
		  }
		 else
		  { v->x() = -sqrt(tmpX/versine);
		  }
		 v->y() = (n.y() + o.x())/(2*v->x()*versine);
		 v->z() = (a.x() + n.z())/(2*v->x()*versine);
		 sine = tmpS / (2*v->x());
		 break;
	       }
	      case Y_MAX:
	       { if (versine < SMALL || tmpY <= 0.0)
		  { *ang = v->y() = v->z() = 0.0;
		    v->x() = 1.0;
		    return;
		  }
		 if ((tmpS = a.x() - n.z()) >= 0.0)
		  { v->y() = sqrt(tmpY/versine);
		  }
		 else
		  { v->y() = -sqrt(tmpY/versine);
		  }
		 v->x() = (n.y() + o.x())/(2*v->y()*versine);
		 v->z() = (o.z() + a.y())/(2*v->y()*versine);
		 sine = tmpS / (2*v->y());
		 break;
	       }
	      case Z_MAX:
	       { if (versine < SMALL || tmpZ <= 0.0)
		  { *ang = v->y() = v->z() = 0.0;
		    v->x() = 1.0;
		    return;
		  }
		 if ((tmpS = n.y() - o.x()) >= 0.0)
		  { v->z() = sqrt(tmpZ/versine);
		  }
		 else
		  { v->z() = -sqrt(tmpZ/versine);
		  }
		 v->x() = (a.x() + n.z())/(2*v->z()*versine);
		 v->y() = (o.z() + a.y())/(2*v->z()*versine);
		 sine = tmpS / (2*v->z());
		 break;
	       }
	    }
	   if (sine < 0)
	    { sine = -sine;
	      v->x() = -v->x();
	      v->y() = -v->y();
	      v->z() = -v->z();
	    }
	 }
	*ang = atan2 (sine, cosine);
}

RxRot& RxRot::mul (const RxRot &m1, const RxRot &m2)
{
	double nx, ox, ax;
	double ny, oy, ay;
	double nz, oz, az;

	nx = m1.n.x() * m2.n.x() + m1.o.x() * m2.n.y() + m1.a.x() * m2.n.z();
	ox = m1.n.x() * m2.o.x() + m1.o.x() * m2.o.y() + m1.a.x() * m2.o.z();
	ax = m1.n.x() * m2.a.x() + m1.o.x() * m2.a.y() + m1.a.x() * m2.a.z();

	ny = m1.n.y() * m2.n.x() + m1.o.y() * m2.n.y() + m1.a.y() * m2.n.z();
	oy = m1.n.y() * m2.o.x() + m1.o.y() * m2.o.y() + m1.a.y() * m2.o.z();
	ay = m1.n.y() * m2.a.x() + m1.o.y() * m2.a.y() + m1.a.y() * m2.a.z();

	nz = m1.n.z() * m2.n.x() + m1.o.z() * m2.n.y() + m1.a.z() * m2.n.z();
	oz = m1.n.z() * m2.o.x() + m1.o.z() * m2.o.y() + m1.a.z() * m2.o.z();
	az = m1.n.z() * m2.a.x() + m1.o.z() * m2.a.y() + m1.a.z() * m2.a.z();

	n.x() = nx;
	n.y() = ny;
	n.z() = nz;

	o.x() = ox;
	o.y() = oy;
	o.z() = oz;

	a.x() = ax;
	a.y() = ay;
	a.z() = az;
	return (*this);
}
     
RxRot& RxRot::mulI (const RxRot &m1, const RxRot &m2)
{
	double nx, ox, ax;
	double ny, oy, ay;
	double nz, oz, az;

	nx = m1.n.x() * m2.n.x() + m1.o.x() * m2.o.x() + m1.a.x() * m2.a.x();
	ox = m1.n.x() * m2.n.y() + m1.o.x() * m2.o.y() + m1.a.x() * m2.a.y();
	ax = m1.n.x() * m2.n.z() + m1.o.x() * m2.o.z() + m1.a.x() * m2.a.z();

	ny = m1.n.y() * m2.n.x() + m1.o.y() * m2.o.x() + m1.a.y() * m2.a.x();
	oy = m1.n.y() * m2.n.y() + m1.o.y() * m2.o.y() + m1.a.y() * m2.a.y();
	ay = m1.n.y() * m2.n.z() + m1.o.y() * m2.o.z() + m1.a.y() * m2.a.z();

	nz = m1.n.z() * m2.n.x() + m1.o.z() * m2.o.x() + m1.a.z() * m2.a.x();
	oz = m1.n.z() * m2.n.y() + m1.o.z() * m2.o.y() + m1.a.z() * m2.a.y();
	az = m1.n.z() * m2.n.z() + m1.o.z() * m2.o.z() + m1.a.z() * m2.a.z();
     
	n.x() = nx;
	n.y() = ny;
	n.z() = nz;

	o.x() = ox;
	o.y() = oy;
	o.z() = oz;

	a.x() = ax;
	a.y() = ay;
	a.z() = az;

	return (*this);
}

RxRot& RxRot::Imul (const RxRot &m1, const RxRot &m2)
{
	double nx, ox, ax;
	double ny, oy, ay;
	double nz, oz, az;

	nx = m2.n.x() * m1.n.x() + m2.n.y() * m1.n.y() + m2.n.z() * m1.n.z();
	ny = m2.n.x() * m1.o.x() + m2.n.y() * m1.o.y() + m2.n.z() * m1.o.z();
	nz = m2.n.x() * m1.a.x() + m2.n.y() * m1.a.y() + m2.n.z() * m1.a.z();

	ox = m2.o.x() * m1.n.x() + m2.o.y() * m1.n.y() + m2.o.z() * m1.n.z();
	oy = m2.o.x() * m1.o.x() + m2.o.y() * m1.o.y() + m2.o.z() * m1.o.z();
	oz = m2.o.x() * m1.a.x() + m2.o.y() * m1.a.y() + m2.o.z() * m1.a.z();

	ax = m2.a.x() * m1.n.x() + m2.a.y() * m1.n.y() + m2.a.z() * m1.n.z();
	ay = m2.a.x() * m1.o.x() + m2.a.y() * m1.o.y() + m2.a.z() * m1.o.z();
	az = m2.a.x() * m1.a.x() + m2.a.y() * m1.a.y() + m2.a.z() * m1.a.z();

	n.x() = nx;
	n.y() = ny;
	n.z() = nz;

	o.x() = ox;
	o.y() = oy;
	o.z() = oz;

	a.x() = ax;
	a.y() = ay;
	a.z() = az;

	return (*this);
}

RxRot& RxRot::ImulI (const RxRot &m1, const RxRot &m2)
{
	double nx, ox, ax;
	double ny, oy, ay;
	double nz, oz, az;

	nx = m1.n.x() * m2.n.x() + m1.n.y() * m2.o.x() + m1.n.z() * m2.a.x();
	ny = m1.o.x() * m2.n.x() + m1.o.y() * m2.o.x() + m1.o.z() * m2.a.x();
	nz = m1.a.x() * m2.n.x() + m1.a.y() * m2.o.x() + m1.a.z() * m2.a.x();

	ox = m1.n.x() * m2.n.y() + m1.n.y() * m2.o.y() + m1.n.z() * m2.a.y();
	oy = m1.o.x() * m2.n.y() + m1.o.y() * m2.o.y() + m1.o.z() * m2.a.y();
	oz = m1.a.x() * m2.n.y() + m1.a.y() * m2.o.y() + m1.a.z() * m2.a.y();

	ax = m1.n.x() * m2.n.z() + m1.n.y() * m2.o.z() + m1.n.z() * m2.a.z();
	ay = m1.o.x() * m2.n.z() + m1.o.y() * m2.o.z() + m1.o.z() * m2.a.z();
	az = m1.a.x() * m2.n.z() + m1.a.y() * m2.o.z() + m1.a.z() * m2.a.z();

	n.x() = nx;
	n.y() = ny;
	n.z() = nz;

	o.x() = ox;
	o.y() = oy;
	o.z() = oz;

	a.x() = ax;
	a.y() = ay;
	a.z() = az;

	return (*this);
}

RxRot& RxRot::slerp (const RxRot &m1, const RxRot &m2, double u)
{
	RxRot mt;
	RxVecd3 v;
	double ang;

	mt.Imul (m1, m2);
	mt.getAxis (&v, &ang);
	mt.setAxis (v, ang*u);
	mul (m1, mt);
	return (*this);
}

RxRot& RxRot::tsp(const RxRot &m1)
{
	double tmp;

	if (&m1 == this)
	 { tmp = o.x();
	   o.x() = n.y();
	   n.y() = tmp;
	   tmp = a.x();
	   a.x() = n.z();
	   n.z() = tmp;
	   tmp = a.y();
	   a.y() = o.z();
	   o.z() = tmp;
	 }
	else
	 { n.x() = m1.n.x();
	   n.y() = m1.o.x();
	   n.z() = m1.a.x();
	   o.x() = m1.n.y();
	   o.y() = m1.o.y();
	   o.z() = m1.a.y();
	   a.x() = m1.n.z();
	   a.y() = m1.o.z();
	   a.z() = m1.a.z();
	 }
	return (*this);
}

RxRot& RxRot::tsp(void)
{
	double tmp;

	tmp = o.x();
	o.x() = n.y();
	n.y() = tmp;
	tmp = a.x();
	a.x() = n.z();
	n.z() = tmp;
	tmp = a.y();
	a.y() = o.z();
	o.z() = tmp;
	return (*this);
}

RxRot& RxRot::setIdent()
{
	n.x() = 1.0; o.x() = 0.0; a.x() = 0.0;
	n.y() = 0.0; o.y() = 1.0; a.y() = 0.0;
	n.z() = 0.0; o.z() = 0.0; a.z() = 1.0;
	return (*this);
}

#if 1
RxVecd3 &RxRot::mul (RxVecd3 &vr, const RxVecd3 &v1) const
{
	register double x, y, z;

	x = v1.x(); y = v1.y(); z = v1.z();
	vr.x() = x * n.x() + y * o.x() + z * a.x();
	vr.y() = x * n.y() + y * o.y() + z * a.y();
	vr.z() = x * n.z() + y * o.z() + z * a.z();
	return (vr);
}

RxVecd3 &RxRot::Imul (RxVecd3 &vr, const RxVecd3 &v1) const
{
	register double x, y, z;

	x = v1.x(); y = v1.y(); z = v1.z();
	vr.x() = x * n.x() + y * n.y() + z * n.z();
	vr.y() = x * o.x() + y * o.y() + z * o.z();
	vr.z() = x * a.x() + y * a.y() + z * a.z();
	return (vr);
}
#endif

RxRot &RxRot::normalize(const RxRot &m1)
{
	/* normalize a transform according to
		
		a = unit(a)
		n = unit(oXa)
		o = aXn
	*/

	a.unit(m1.a);
	n.cross (m1.o, a);
	n.unit ();
	o.cross (a, n);
	return (*this);
}

int RxRot::oscan (RxParseObj *obj)
{
	RxParseObj *lobj;
	int len;

	len = obj->getListLength();
	if (!obj->isNumberList() || (len != 9 && len != 4))
	 { RxSysError (RxSysError::ParsingError,
"RxRot::oscan: object not a list of 4 or 9 numbers\n").post();
	   return 0;
	 }
	if (len == 9)
	 { return RxMatdBase::oscan (obj);
	 }
	else /* len == 4 */
	 { RxVecd3 axis;
	   double ang;

	   lobj = obj->firstChild();
	   axis.x() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	   axis.y() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	   axis.z() = lobj->getNumberValue();

	   lobj = obj->nextChild();
	   ang = lobj->getNumberValue();
	   
	   setAxisD (axis, ang);
	   return 1;
	 }
}

int RxRot::cprint (RxCoutput *co, RxPrintSpec *ps) const
{
	if (ps->findSpecial ("%M"))
	 { return RxMatdBase::cprint (co, ps);
	 }
	else
	 { const RxFmtNode *cf0;
	   const char *fmtT;	// format for translational component
	   const char *fmtR;	// format for rotational component
	   RxVecd3 axis;
	   double ang;

	   // figure out format strings
	   cf0 = ps->convNode(0);
	   if (cf0)
	    { fmtT = cf0->fmt();
	    }
	   else
	    { fmtT = "%8.3f";
	    }
	   if (cf0 && ps->convNode(1))
	    { fmtR = ps->convNode(1)->fmt();
	    }
	   else
	    { fmtR = fmtT;
	    }
	   co->output_str (ps->prefix());
	   co->output_char ('[');
	   getAxisD (&axis, &ang);
	   co->output_doubleArray (fmtR, NULL, NULL, axis.buf(), 3);
	   co->output_char (' ');
	   co->output_double (fmtT, ang);
	   co->output_char (']');
	   if (ps->numConvs() == 0)
	    { co->output_char ('\n');
	    }
	   else
	    { co->output_str (ps->suffix());
	    }
	   return co->nchars();
	 }
}
