#include <RxVecdSpline.h>
#include <cdefs.h>

RxVecdSpline::RxVecdSpline (int len, int deg, int rat)
{
	degree = MAX(1,deg);
	rational = rat;
	vlen = MAX(1,len);
	lastLeftKnot = NULL;
	headIdx = 0;
}

void RxVecdSpline::reset ()
{
	knots.removeAndDeleteAll();
	headIdx = 0;
}

void RxVecdSpline::reset (int deg, int rat)
{
	reset();
	degree = MAX(1,deg);
	rational = rat;
	lastLeftKnot = NULL;
}

RxVecdSknot *RxVecdSpline::findLeftKnot (double u, RxVecdSknot *start)
{
	// Find the highest knot kl such that u is contained in the
	// interval [kl->u, kl->next()->u) 

	RxVecdSknot *k0;

	if ((k0 = start) == NULL)
	 { k0 = knots.head();
	 }
	while (k0 && k0->u <= u)
	 { k0 = k0->next();
	 }
	while (k0 && k0->u > u)
	 { k0 = k0->prev();
	 }
	return (k0);
}

RxVecdSknot *RxVecdSpline::findRightKnot (double u, RxVecdSknot *start)
{
	// Find the lowest knot kr such that u is contained in the
	// interval (kr->prev()->u, kr->u]

	RxVecdSknot *k0;

	if ((k0 = start) == NULL)
	 { k0 = knots.tail();
	 }
	while (k0 && k0->u >= u)
	 { k0 = k0->prev();
	 }
	while (k0 && k0->u < u)
	 { k0 = k0->next();
	 }
	return (k0);
}

void RxVecdSpline::deBoor (RxVecd *v, RxVecd *dv, RxVecd *ddv, 
	RxVecdSknot *k0, RxVecdSknot *k1, double u, int k)
{
	if (k == 1)
	 { if (v)
	    { v->scale ((k1->u-u)/(k1->u-k0->u), *k0->v);
	      v->extrap (*v, (u-k0->u)/(k1->u-k0->u), *k0->next()->v);
	    }
	   if (dv)
	    { dv->sub (*k0->next()->v, *k0->v);
	      dv->scale (1/(k1->u-k0->u));
	    }
	   if (ddv)
	    { ddv->setZero();
	    }
	 }
	else
	 { RxVecd p0(vlen);
	   RxVecd p1(vlen);
	   RxVecd d0(vlen);
	   RxVecd d1(vlen);

	   deBoor (&p0, ddv ? &d0 : NULL, NULL, 
		   k0->prev(), k1, u, k-1);
	   deBoor (&p1, ddv ? &d1 : NULL, NULL, 
		   k0, k1->next() ? k1->next() : k1, u, k-1);
	   if (v)
	    { v->scale ((k1->u-u)/(k1->u-k0->u), p0);
	      v->extrap (*v, (u-k0->u)/(k1->u-k0->u), p1);
	    }
	   if (dv)
	    { dv->sub (p1, p0);
	      dv->scale (k/(k1->u-k0->u));
	    }
	   if (ddv)
	    { ddv->sub (d1, d0);
	      ddv->scale (k/(k1->u-k0->u));
	    }
	 }
}

void RxVecdSpline::addPnt (RxVecd *v, double u)
{
	RxVecdSknot *kn = new RxVecdSknot();
	RxVecdSknot *kt = knots.tail();

	kn->v = v;
	kn->u = u;
	kn->w = 1;
	if (kt)
	 { kn->idx = kt->idx+1;
	 }
	else
	 { kn->idx = headIdx;
	 }
	if (kt && kt->u == u)
	 { kn->m = kt->m+1;
	   do
	    { kt->m = kn->m;
	      kt = kt->prev();
	    }
	   while (kt && kt->u == u);
	 }
	else
	 { kn->m = 1;
	 }
	knots.addTail (kn);
}

void RxVecdSpline::addPnt (RxVecd *v, double w, double u)
{
	RxVecdSknot *kn = new RxVecdSknot();
	RxVecdSknot *kt = knots.tail();

	kn->v = v;
	kn->u = u;
	kn->w = w;
	kn->m = 1;
	if (kt)
	 { kn->idx = kt->idx+1;
	 }
	else
	 { kn->idx = headIdx;
	 }
	if (kt && kt->u == u)
	 { kn->m = kt->m+1;
	   do
	    { kt->m = kn->m;
	      kt = kt->prev();
	    }
	   while (kt && kt->u == u);
	 }
	else
	 { kn->m = 1;
	 }
	knots.addTail (kn);
}

void RxVecdSpline::removePnts (double u)
{
	RxVecdSknot *kn;
	RxVecdSknot *kx;
	int k;
	int m = 1;

	kn = findLeftKnot (u, lastLeftKnot);
	if (lastLeftKnot && lastLeftKnot->u <= u)
	 { lastLeftKnot = NULL;
	 }
	for (k=degree-1; k-- && kn; kn=kn->prev())
	   ;
	if (kn && kn->prev())
	 { knots.removeAndDelete (knots.head(), kn->prev());
	   // reset head knot multiplicty
	   for (kx=kn->next(); kx && kx->u==kn->u; kx=kx->next())
	    { m++;
	    }
	   kn->m = m;
	   for (kx=kn->next(); kx && kx->u==kn->u; kx=kx->next())
	    { kx->m = m;
	    }
	   if ((kn = knots.head()))
	    { headIdx = kn->idx;
	    }
	   else
	    { headIdx = 0;
	    }
	 }
}

int RxVecdSpline::calcPnt (RxVecd *v, double u)
{
	RxVecdSknot *kl;
	RxVecdSknot *kr;
	RxVecdSknot *kt;

	kl = findLeftKnot (u, lastLeftKnot);
	if (kl == NULL)
	 { kt = knots.tail();
	   if (kt && u == kt->u && kt->prev())
	    { kl = kt->prev();
	    }
	   else
	    { return (-1);
	    }
	 }
	// make sure there are degree-1 knots left of kl 
	if (kl && kl->idx - headIdx < degree-1)
	 { return (-1);
	 }
	kr = kl->next();
	deBoor (v, NULL, NULL, kl, kr, u, degree);
	return (1);	
}

int RxVecdSpline::calcPnt (RxVecd *v, RxVecd *dv, RxVecd *ddv,
			    double u, int side)
{
	RxVecdSknot *kl;
	RxVecdSknot *kr;
	RxVecdSknot *kt;

	if (side == RX_SPLINE_LEFT)
	 { kr = findRightKnot (u, lastLeftKnot);
	   if (kr == NULL)
	    { return -1;
	    }
	   kl = kr->prev();
	 }
	else
	 { kl = findLeftKnot (u, lastLeftKnot);
	   if (kl == NULL)
	    { kt = knots.tail();
	      if (!dv && !ddv && kt && u == kt->u && kt->prev())
	       { kl = kt->prev();
	       }
	      else
	       { return -1;
	       }
	    }
	   kr = kl->next();
	 }
	// make sure there are degree-1 knots left of kl 
	if (kl->idx - headIdx < degree-1)
	 { return -1;
	 }
	deBoor (v, dv, ddv, kl, kr, u, degree);
	return (1);
}

int RxVecdSpline::numKnots ()
{
	RxVecdSknot *kt;

	if ((kt = knots.tail()))
	 { return (kt->idx - headIdx + 1);
	 }
	else
	 { return (0);
	 }
}

RxVecd *RxVecdSpline::firstPnt (double *wp, double *up, int *mp)
{
	curPnt = knots.head();
	if (curPnt)
	 { if (up) *up = curPnt->u;
	   if (wp) *wp = curPnt->w;
	   if (mp) *mp = curPnt->m;
	   return (curPnt->v);
	 }
	else
	 { return (NULL);
	 }
}

RxVecd *RxVecdSpline::nextPnt (double *wp, double *up, int *mp)
{
	if (curPnt)
	 { curPnt = curPnt->next();
	 }
	if (curPnt)
	 { if (up) *up = curPnt->u;
	   if (wp) *wp = curPnt->w;
	   if (mp) *mp = curPnt->m;
	   return (curPnt->v);
	 }
	else
	 { return (NULL);
	 }
}
