#include <stdio.h>
#include <cdefs.h>
#include <assert.h>

#include <RxVecdSpline.h>
#ifdef HAS_GRF4
#include <grf4.h>
#endif

char **argvec;
int interactive = 0;

#ifdef HAS_GRF4
int winw = 512;
int winh = 512;
double lastx = -1000;
double lasty = -1000;
double scaleX;
double scaleY;
int degree = 3;
int knotnum = 0;

RxVecdSpline *spline;
RxVecd *dragV = NULL;
char *expStr;

#define WIN_VSIZE	100.0
#define KNOT_WIN_SIZE	5
#define EPS	(1e-11)

typedef struct {
	double u, x, y;
} TestRec;

TestRec testList[] = {
	0,   0,  0, 
	0,  25,  0, 
	0,  25, 25, 
	1,   0, 25, 
	2, -25, 25, 
	3, -25,  0, 
	3, -25,  0, 
       -1,   0,  0, 
	4,   0,-25,
	4,  25,  0,
	4,  25, 25,
	5,  50, 25,
       -2,   0,  0
};

int doSplineTest (RxVecdSpline *spline);

void drawX (double x, double y, double size)
{
	grf4Move2 (x-size, y-size);
	grf4Draw2 (x+size, y+size);
	grf4Move2 (x-size, y+size);
	grf4Draw2 (x+size, y-size);
}

void doDraw(void *)
{
	RxVecd *v;
	RxVecd vec(2);
	RxVecd dv(2);
	RxVecd ddv(2);
	double u;
	int i, i0;

	grf4Clear();
	for (v=spline->firstPnt(); v; v=spline->nextPnt())
	 { drawX ((*v)(0), (*v)(1), 2.0);
	 }
	if (knotnum > 0)
	 { 
#define NUM_INTERVALS	20
	   i0 = MAX(0,knotnum-KNOT_WIN_SIZE)*NUM_INTERVALS;
	   for (i=i0; i<=NUM_INTERVALS*knotnum; i++)
	    { if (spline->calcPnt (&vec, &dv, &ddv, 
				   i/(double)NUM_INTERVALS) == -1)
	       { printf ("ERROR, io=%d u=%g\n", i0,
			 i/(double)NUM_INTERVALS);
	       }
	      if (i==i0)
	       { grf4Move2 (vec(0), vec(1));
	       }
	      else
	       { grf4Draw2 (vec(0), vec(1));
	       }
	    }
	 }
	grf4Flush();
}

getCursor (double *xp, double *yp)
{	
	Grf4Event evx;

	grf4PollEvent (GRF_CURSOR, &evx);
	*xp = (evx.val1-winw/2)*scaleX;
	*yp = (evx.val2-winh/2)*scaleY;
}

void doResize (int w, int h)
{
	winw = w;
	winh = h;
	scaleX = WIN_VSIZE/MIN(winw,winh);
	scaleY = scaleX;
	grf4Ortho (-scaleX*winw/2, scaleX*winw/2, 
		   -scaleY*winh/2, scaleY*winh/2, 
		   -1.0, 1.0, /*aspectRatio=*/1);
}

void onResize (Grf4Event *ev, int)
{
	doResize (ev->val1, ev->val2);
}

void onMotion (Grf4Event *ev, int)
{
	double x, y;

	if (dragV)
	 { getCursor (&x, &y);
	   (*dragV)(0) = x;
	   (*dragV)(1) = y;
	   doDraw(0);
	 }
}

void onButton (Grf4Event *ev, int)
{
	switch (ev->code)
	 { case GRF_RIGHT_MOUSE:
	    { 
	      break;
	    }
	   case GRF_LEFT_MOUSE:
	    { if (ev->val1)
	       { /* button pressed */
		 RxVecd *v = new RxVecd(2);
		 int k;

		 getCursor (&lastx, &lasty);
		 (*v)(0) = lastx;
		 (*v)(1) = lasty;
		 knotnum++;
		 spline->addPnt (v, knotnum);
		 if (knotnum > KNOT_WIN_SIZE)
		  { spline->removePnts (knotnum-KNOT_WIN_SIZE);
		  }
		 dragV = v;
		 doDraw(0);
	       }
	      else
	       { dragV = NULL;
	       }
	      break;
	    }
	 }
}

#endif

int doBatchTests()
{
	RxVecd *v;
	int line;
	int i;
	int deg;

	for (deg=0; deg<=4; deg++)
	 { spline->reset (deg, 0);
	   for (i=0; i<ARRAY_SIZE(testList); i++)
	    { if (testList[i].u >= 0)
	       { v = new RxVecd(spline->getVlength());
		 v->setZero();
		 (*v)(0) = testList[i].x;
		 (*v)(1) = testList[i].y;
		 spline->addPnt (v, testList[i].u);
	       }
	      else
	       { spline->removePnts (-testList[i].u);
	       }
	      printf ("numKnots=%d\n", spline->numKnots());
	      line = doSplineTest(spline);
	      if (line != 0)
	       { fprintf (stderr, "Error, i=%d, line %d:\n%s\n",
				   i, line, expStr);
		 exit (1);
	       }
	    }
	 }
	printf ("\nROUTINES OK\n\n");
}

main(int argc, char **argv)
{
	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT("-i"))
	    { interactive = 1;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [-i]\n", argvec[0]);
	   exit (1);
	 }
	spline = new RxVecdSpline (2, degree, 0);

	if (!interactive)
	 { doBatchTests();
	 }
	else
	 { 
#ifndef HAS_GRF4
	   printf ("Need GRF4 graphics to support -i option\n");
	   exit (0);
#else
	   Grf4Event evx;
	   RxVecd *v;

	   grf4PlaceWindow (200, 400, winw, winh);
	   grf4Init ("DAO", GRF_WHITE, GRF_BLACK, GRF_2DONLY | GRF_X11);

	   grf4PollEvent (GRF_RESIZE, &evx);
	   doResize (evx.val1, evx.val2);

	   grf4AttachEvent (GRF_RESIZE_EVENT, onResize, 0);
	   grf4AttachEvent (GRF_MOTION_EVENT, onMotion, 0);
	   grf4AttachEvent (GRF_BUTTON_EVENT, onButton, 0);

	   grf4SetDrawFxn (doDraw, (void*)0);
	
	   v = new RxVecd(2);
	   (*v)(0) = 0; (*v)(1) = 0;
	   spline->addPnt (&(*v), 0);
	   v = new RxVecd(2);
	   (*v)(0) = 25; (*v)(1) = 0;
	   spline->addPnt (&(*v), 0);
	   v = new RxVecd(2);
	   (*v)(0) = 25; (*v)(1) = 25;
	   spline->addPnt (&(*v), 0);
	   v = new RxVecd(2);
	   (*v)(0) = 0; (*v)(1) = 25;
	   spline->addPnt (&(*v), 1);
	   v = new RxVecd(2);
	   (*v)(0) = -25; (*v)(1) = 25;
	   spline->addPnt (&(*v), 2);
	   knotnum = 2;

	   doDraw(0);
	   grf4Wait (-1);
#endif
	 }
	return 0;
}

#define CHECK(EX) if (!(EX)) { expStr = #EX; return (__LINE__); }

int doSplineTest (RxVecdSpline *spline)
{
	RxVecd vec0(spline->getVlength());
	RxVecd dv0(spline->getVlength());
	RxVecd ddv0(spline->getVlength());

	RxVecd vec1(spline->getVlength());
	RxVecd dv1(spline->getVlength());
	RxVecd ddv1(spline->getVlength());

	RxVecd a5(spline->getVlength());
	RxVecd a4(spline->getVlength());
	RxVecd a3(spline->getVlength());
	RxVecd a2(spline->getVlength());
	RxVecd a1(spline->getVlength());
	RxVecd a0(spline->getVlength());

	RxVecd *vr, *vl = NULL;
	double u, ur, ul, h;
	int mr, ml;
	int k, i, j;

	// spline consistency checks

	vr = spline->firstPnt (NULL, &ur, &mr);
	k = 0;

	do
	 { vl = vr;
	   ul = ur;
	   ml = mr;
	   vr = spline->nextPnt (NULL, &ur, &mr);
	   k++;
	   if (vr && ur != ul)
	    { // there is an interval here
	      if (k >= spline->getDegree())
	       { if (k+mr >= spline->numKnots())
		  { CHECK (spline->calcPnt (&vec0, &dv0, &ddv0, ur,
					  RX_SPLINE_RIGHT) == -1);
		  }
		 else
		  { CHECK (spline->calcPnt (&vec0, &dv0, &ddv0, ur,
					  RX_SPLINE_RIGHT) == 1);
       		  }
		 if (k-ml >= spline->getDegree())
		  { CHECK (spline->calcPnt (&vec0,&dv0,&ddv0,ul,
					  RX_SPLINE_LEFT) == 1);
		  }
		 else
		  { CHECK (spline->calcPnt (&vec0,&dv0,&ddv0,ul,
					  RX_SPLINE_LEFT) == -1);
		  }
		 CHECK (spline->calcPnt (&vec0, &dv0, &ddv0, ul,
					  RX_SPLINE_RIGHT) ==1);
		 CHECK (spline->calcPnt (&vec1, &dv1, &ddv1, ur,
					  RX_SPLINE_LEFT)==1);
		 dv0.scale (ur-ul);
		 dv1.scale (ur-ul);
		 ddv0.scale ((ur-ul)*(ur-ul));
		 ddv1.scale ((ur-ul)*(ur-ul));
		 if (spline->getDegree() <= 3)
		  { for (i=0; i<spline->getVlength(); i++)
		     { a3(i) = dv1(i)-2*vec1(i)+dv0(i)+2*vec0(i);
		       a2(i) =-dv1(i)+3*vec1(i)-2*dv0(i)-3*vec0(i);
		       a1(i) = dv0(i);
		       a0(i) = vec0(i);
		     }
		  }
		 else
		  { for (i=0; i<spline->getVlength(); i++)
		     { a5(i) =  .5*ddv1(i) - 3*dv1(i) + 6*vec1(i) -
				.5*ddv0(i) - 3*dv0(i) - 6*vec0(i);
		       a4(i) =    -ddv1(i) + 7*dv1(i) - 15*vec1(i) +
			       1.5*ddv0(i) + 8*dv0(i) + 15*vec0(i);
		       a3(i) =  .5*ddv1(i) - 4*dv1(i) + 10*vec1(i) -
			       1.5*ddv0(i) - 6*dv0(i) - 10*vec0(i);
		       a2(i) =  .5*ddv0(i);
		       a1(i) =     dv0(i);
		       a0(i) =     vec0(i);
		     }
		  }
		 for (j=0; j<=NUM_INTERVALS; j++)
		  { h = j/(double)NUM_INTERVALS;
		    u = (1-h)*ul + h*ur;
		    
		    if (spline->getDegree() <= 3)
		     { for (i=0; i<spline->getVlength(); i++)
			{ vec0(i) = ((a3(i)*h+a2(i))*h+a1(i))*h+a0(i);
			  dv0(i) = (3*a3(i)*h+2*a2(i))*h+a1(i);
			  ddv0(i) = 6*a3(i)*h+2*a2(i);
			}
		     }
		    else
		     { for (i=0; i<spline->getVlength(); i++)
			{ vec0(i) = ((((a5(i)*h + a4(i))*h + a3(i))*h +
				       a2(i))*h+a1(i))*h+a0(i);
			  dv0(i) =  (((5*a5(i)*h + 4*a4(i))*h + 3*a3(i))*h +
				       2*a2(i))*h+a1(i);
			  ddv0(i) = ((20*a5(i)*h + 12*a4(i))*h + 6*a3(i))*h +
				       2*a2(i);
			}
		     }
		    if (j<NUM_INTERVALS)
		     { CHECK (spline->calcPnt (&vec1, &dv1, &ddv1,
					u, RX_SPLINE_RIGHT) == 1);
		     }
		    else
		     { CHECK (spline->calcPnt (&vec1, &dv1, &ddv1,
					u, RX_SPLINE_LEFT) == 1);
		     }
		    CHECK (vec1.equal(vec0, EPS));
		    CHECK (dv1.equal(dv0, EPS));
		    CHECK (ddv1.equal(ddv0, EPS));
		  }
	       }
	      else
	       { CHECK (spline->calcPnt (&vec0, ul) == -1);
		 CHECK (spline->calcPnt (&vec0, (ul+ur)/2) == -1);
		 CHECK (spline->calcPnt (&vec0, &dv0, &ddv0, ur,
					  RX_SPLINE_LEFT) == -1);
		 if (k >= spline->getDegree()-1 && k+1 < spline->numKnots())
		  { CHECK (spline->calcPnt (&vec0, &dv0, &ddv0, ur,
			  		    RX_SPLINE_RIGHT) == 1);
		  }
		 else
		  { CHECK (spline->calcPnt (&vec0, &dv0, &ddv0, ur,
					    RX_SPLINE_RIGHT) == -1);
		  }
	       }
	    }
	 }
	while (vr);
	CHECK (k == spline->numKnots());
	return (0);
}
