#include <grf4.h>
#include <math.h>
#include <mathx.h>
#include "grf4Support.h"

typedef struct {
	TRSF T;
	double size;
	int enabled;

#define JACK_X		0x01
#define JACK_Y		0x02
#define JACK_Z		0x04
#define JACK_XYZ	0x07
#define JACK_TRANS_XY	JACK_Z
#define JACK_TRANS_YZ	JACK_X
#define JACK_TRANS_ZX	JACK_Y
#define JACK_ROT	0x10
#define JACK_ROT_X	(JACK_ROT|JACK_X)
#define JACK_ROT_Y	(JACK_ROT|JACK_Y)
#define JACK_ROT_Z	(JACK_ROT|JACK_Z)
	int select;

	VECT nrm;
	double off;
	VECT p0;
	TRSF T0;
} Jack;

static int jackInitDone = 0;
static Jack jack;

static void *whiteMat;
static void *redMat;

static VECT *jackCodeToAxis (Jack *jack, int code)
{
	switch (code)
	 { case JACK_X:
	    { return &jack->T.n;
	    }
	   case JACK_Y:
	    { return &jack->T.o;
	    }
	   case JACK_Z:
	    { return &jack->T.a;
	    }
	   default:
	    { fprintf (stderr,
"Internal error: jackCodeToAxis(): bogus axis code %x\n", code);
	      exit (1);
	    }
	 }
}

static void jackCodeToLocalAxis (Jack *jack, VECT *axis, int code)
{
	zeroVect (axis);
	switch (code)
	 { case JACK_X:
	    { axis->x = 1;
	      break;
	    }
	   case JACK_Y:
	    { axis->y = 1;
	      break;
	    }
	   case JACK_Z:
	    { axis->z = 1;
	      break;
	    }
	   default:
	    { fprintf (stderr,
"Internal error: jackCodeToLocalAxis(): bogus axis code %x\n", code);
	      exit (1);
	    }
	 }
}

static int jackAxisToCode (Jack *jack, VECT *u)
{
	if (u == &jack->T.n)
	 { return JACK_X;
	 }
	else if (u == &jack->T.o)
	 { return JACK_Y;
	 }
	else if (u == &jack->T.a)
	 { return JACK_Z;
	 }
	else
	 { fprintf (stderr,
"Internal error: jackAxisToCode(): axis not contained in jack->T\n");
	   exit (1);
	 }
	return -1;
}

static void jackInit(Jack *jack, double s)
{
	identTrsf (&jack->T);
	jack->size = s;
	jack->select = 0;
	jack->enabled = 0;
	whiteMat = grf4DefMaterial(0xffffff, 0, 0xffffff, 0, 0.0, 1.0);
	redMat = grf4DefMaterial(0xff, 0, 0xff, 0, 0.0, 1.0);
}

static void jackDrawAxis (Jack *jack, int code_u, int code_v)
{
	VECT u;
	VECT v;
	double a = jack->size;
	int highliteU = 0;
	int highliteV = 0;

	jackCodeToLocalAxis (jack, &u, code_u);
	jackCodeToLocalAxis (jack, &v, code_v);

	if (jack->select)
	 { if (jack->select & JACK_ROT)
	    { if ((jack->select|code_u|code_v) == (JACK_ROT|JACK_XYZ))
	       { highliteU = 1;
		 highliteV = 1;
	       }
	    }
	   else
	    { if (jack->select != code_u)
	       { highliteU = 1;
	       }
	    }
	 }

	grf4SetMaterial (highliteU ? redMat : whiteMat);
	grf4Move (-a*u.x, -a*u.y, -a*u.z);
	grf4Draw ( a*u.x,  a*u.y,  a*u.z);
	grf4SetMaterial (highliteV ? redMat : whiteMat);
	grf4Move (-a*(u.x+v.x/3), -a*(u.y+v.y/3), -a*(u.z+v.z/3));
	grf4Draw (-a*(u.x-v.x/3), -a*(u.y-v.y/3), -a*(u.z-v.z/3));
	grf4Move ( a*(u.x+v.x/3),  a*(u.y+v.y/3),  a*(u.z+v.z/3));
	grf4Draw ( a*(u.x-v.x/3),  a*(u.y-v.y/3),  a*(u.z-v.z/3));
}

static void jackDraw (Jack *jack)
{
	double a = jack->size;
	grf4PushTrans();
	grf4MultTrans (&jack->T);
	jackDrawAxis (jack, JACK_X, JACK_Y);
	jackDrawAxis (jack, JACK_Y, JACK_Z);
	jackDrawAxis (jack, JACK_Z, JACK_X);
	grf4PopTrans();
}

static double jackDistToSeg (VECT *u, VECT *b, int sx, int sy, double halfLen)
{
	VECT p0, p1, dp;
	double lam;

	lam = grf4ScrToLine (&p0, sx, sy, u, b);
	if (lam == HUGE_VAL)
	 { return lam;
	 }
	else
	 { if (lam > halfLen)
	    { lam = halfLen;
	    }
	   if (lam < -halfLen)
	    { lam = -halfLen;
	    }
	   combineVect (&p1, lam, u, 1.0, b);
	   subtractVect (&dp, &p0, &p1);
	   return norm2Vect (&dp);
	 }
}

static double jackCheckAxis (Jack *jack, int code_u, int code_v,
		      int sx, int sy, double min_d)
{
	VECT b;
	VECT *u;
	VECT *v;
	double d;

	u = jackCodeToAxis (jack, code_u);
	v = jackCodeToAxis (jack, code_v);
	b = jack->T.p;
	d = jackDistToSeg (u, &b, sx, sy, jack->size);
printf ("u %d dist=%g\n", code_u, d);
	if (d < min_d)
	 { min_d = d;
	   if (d < 0.05*jack->size)
	    { jack->select = code_u;
	    }
	 }
	combineVect (&b, 1.0, &jack->T.p, jack->size, u);
	d = jackDistToSeg (v, &b, sx, sy, jack->size/3);
printf ("v %d dist=%g\n", code_v, d);
	if (d < min_d)
	 { min_d = d;
	   if (d < 0.05*jack->size)
	    { jack->select = JACK_ROT | (JACK_XYZ & ~(code_u | code_v));
	    }
	 }
	combineVect (&b, 1.0, &jack->T.p, -jack->size, u);
	d = jackDistToSeg (v, &b, sx, sy, jack->size/3);
printf ("v %d dist=%g\n", code_v, d);
	if (d < min_d)
	 { min_d = d;
	   if (d < 0.05*jack->size)
	    { jack->select = JACK_ROT | (JACK_XYZ & ~(code_u | code_v));
	    }
	 }
	return min_d;
}

static void jackCheck (Jack *jack)
{
	Grf4Event curs;
	int sx, sy;
	double min_d;

	/* get current cursor position */
	grf4PollEvent (GRF_CURSOR, &curs);
	sx = curs.val1;
	sy = curs.val2;

	jack->select = 0;
	min_d = jackCheckAxis (jack, JACK_X, JACK_Y, sx, sy, HUGE_VAL);
	min_d = jackCheckAxis (jack, JACK_Y, JACK_Z, sx, sy, min_d);
	min_d = jackCheckAxis (jack, JACK_Z, JACK_X, sx, sy, min_d);
printf ("\n");
	if (jack->select)
	 { jack->nrm = *jackCodeToAxis(jack, jack->select & JACK_XYZ);
	   jack->off = -dotVect (&jack->nrm, &jack->T.p);
	   jack->T0 = jack->T;
	   grf4ScrToPlane (&jack->p0, sx, sy, &jack->nrm, jack->off);
	 }
}

void grf4JackOnCursor(Grf4Event *ev, void *ptr, int arg)
{
	Jack *jackp = &jack;

	if (jackp->enabled && jackp->select)
	 { VECT p1;
	   VECT dp;

	   grf4ScrToPlane (&p1, ev->val1, ev->val2, &jackp->nrm, jackp->off);
	   subtractVect (&dp, &p1, &jackp->p0);
	   if (jackp->select & JACK_ROT)
	    { VECT dp0;
	      VECT dp1;
	      VECT dpx;
	      VECT axis;
	      double s, c;

	      subtractVect (&dp0, &jackp->p0, &jackp->T0.p);
	      subtractVect (&dp1, &p1, &jackp->T0.p);
	      unitVect (&dp0, &dp0);
	      unitVect (&dp1, &dp1);
	      c = dotVect (&dp0, &dp1);
	      crossVect (&dpx, &dp0, &dp1);
	      s = dotVect (&dpx, &jackp->nrm);
	      jackp->T = jackp->T0;
	      jackCodeToLocalAxis (jackp, &axis, jackp->select & JACK_XYZ);
	      multTrsfRot (&jackp->T, &axis, RTOD*atan2 (s, c));
	    }
	   else
	    { VECT dp;
	      subtractVect (&dp, &p1, &jackp->p0);
	      addVect (&jackp->T.p, &jackp->T0.p, &dp);
	    }
	 }
}

void grf4JackOnButton (Grf4Event *ev, void *ptr, int arg)
{
	Jack *jackp = &jack;

	if (jackp->enabled)
	 { if (ev->code == GRF_RIGHT_MOUSE)
	    { if (ev->val1)
	       { jackCheck (jackp);
	       }
	      else
	       { jackp->select = 0;
	       }
	    }
	 }
}

void grf4EnableJack (int enable)
{
	if (jackInitDone == 0)
	 { jackInit (&jack, 1.0);
	   jackInitDone = 1;
	 }
	jack.enabled = enable;
}

void grf4SetJackTrsf (TRSF *T)
{
	if (jackInitDone == 0)
	 { jackInit (&jack, 1.0);
	   jackInitDone = 1;
	 }
	jack.T = *T;
}

void grf4GetJackTrsf (TRSF *T)
{
	if (jackInitDone == 0)
	 { jackInit (&jack, 1.0);
	   jackInitDone = 1;
	 }
	*T = jack.T;
}

void grf4SetJackSize (double size)
{
	if (jackInitDone == 0)
	 { jackInit (&jack, 1.0);
	   jackInitDone = 1;
	 }
	jack.size = size;
}

void grf4DrawJack ()
{
	jackDraw (&jack);
}

TRSF *grf4GetJackTrsfPtr()
{
	return &jack.T;
}

