#include <RxMatdBase.h>
#include <RxMathError.h>
#include <math.h>
#include <cdefs.h>
#include <stdlib.h>
#include <RxPrintSpec.h>
#include <RxCio.h>
#include <RxParseObj.h>
#include <RxSysError.h>

#define EPSILON	(1e-15)

int RxMatdBase::doInvert(const RxMatdBase &m1)
{
	int n, i, j, k, imax;
	double *ar;
	double *ai;
	double det;
	double max;
	double val;

	if (nrows() != ncols())
	 { RxMathError(RxMathError::BadSize, 
	      "RxMatdBase::invert(): matrix not square\n").raise();
	   return (0);
	 }

	n = nrows();
	ai = (double *) malloc (n * n * sizeof(double));
	ar = (double *) malloc (n * n * sizeof(double));

	det = 1.0;
	for (i=0; i<n; i++)
	 { for (j=0; j<n; j++)
	    { if (i == j)
	       { ar [i*n+j] = 1.0;
	       }
	      else
	       { ar [i*n+j] = 0.0;
	       }
	      ai[i*n+j] = m1(i,j);
	    }
	 }

	for (k=0; k<n; k++)
	 {
	   /* find max element in column k */

	   imax = k;
	   max = ai[imax*n+k];
	   for (i=k; i<n; i++)
	    { if (ABS(ai[i*n+k]) > max)
	       { imax = i;
		 max = ABS(ai[imax*n+k]);
	       }
	    }

	   val = ai[imax*n+k];

	   if (imax != k)
	    { /* interchange rows in both matrices */

	      double temp;

	      for (j=0; j<n; j++)
	       { temp = ai[imax*n+j];
		 ai[imax*n+j] = ai[k*n+j];
		 ai[k*n+j] = temp;
	       }
	      for (j=0; j<n; j++)
	       { temp = ar[imax*n+j];
		 ar[imax*n+j] = ar[k*n+j];
		 ar[k*n+j] = temp;
	       }
	      det = -det;
	    }
	   
	   if (max < EPSILON)
	    { RxMathError (RxMathError::Singular, 
			  "RxMatdBase::invert(): %K\n").raise();
	      free (ai);
	      free (ar);
	      return (0);
	    }

	   /* reduce row k */
	  
	   for (j=0; j<n; j++)
	    { ai[k*n+j] /= val;
	      ar[k*n+j] /= val;
	    }
	   det *= val;
	   
	   /* reduce rows other than k */

	   for (i=0; i<n; i++)
	    { if (i != k)
	       { double mult = ai[i*n+k];
		 for (j=0; j<n; j++)
		  { ai[i*n+j] -= mult*ai[k*n+j];
		    ar[i*n+j] -= mult*ar[k*n+j];
		  }
	       }
	    }
	 }

	for (i=0; i<n; i++)
	 { for (j=0; j<n; j++)
	    { (*this)(i,j) = ar[i*n+j];
	    }
	 }
	free (ai);
	free (ar);
	return (1);
}

double RxMatdBase::dum;

double &RxMatdBase::elem(int i, int j)
{
	static double dum;

	if (i < 0 || i >= nrows() || j < 0 || j >= ncols())
	 { RxMathError(RxMathError::BadIndex, 
		       "RxMatdBase::elem: %K (%d,%d)\n", i,j).raise();
	   return (dum);
	 }
	else
	 { return ((*this)(i,j));
	 }
}

double RxMatdBase::elem(int i, int j) const
{
	static double dum;

	if (i < 0 || i >= nrows() || j < 0 || j >= ncols())
	 { RxMathError(RxMathError::BadIndex, 
		       "RxMatdBase::elem: %K (%d,%d)\n", i,j).raise();
	   return (dum);
	 }
	else
	 { return ((*this)(i,j));
	 }
}

#if 0
int RxMatdBase::scan (FILE *fp)
{
	int i, j;
	int status;

	for (i=0; i<nrows(); i++)
	 { for (j=0; j<ncols(); j++)
	    { if ((status = fscanf (fp, "%lf", &(*this)(i,j))) != 1)
	       { break;
	       }
	    }
	 }
	return ((i==nrows() && j==ncols()) ? 1 : status );
}

int RxMatdBase::xscan (FILE *fp)
{
	int i, j;
	int status;

	for (i=0; i<nrows(); i++)
	 { for (j=0; j<ncols(); j++)
	    { if ((status = xscanf (fp, "%lf", &(*this)(i,j))) != 1)
	       { break;
	       }
	    }
	 }
	return ((i==nrows() && j==ncols()) ? 1 : status );
}

int RxMatdBase::scan (const char *buf, int *cntp)
{
	int i, j;
	int status;
	int idx = 0;
	int cnt;

	for (i=0; i<nrows(); i++)
	 { for (j=0; j<ncols(); j++)
	    { status = sscanf (&buf[idx], "%lf%n", &(*this)(i,j), &cnt);
	      if (status != 1)
	       { break;
	       }
	      idx += cnt;
	    }
	 }
	if (cntp)
	 { *cntp = idx;
	 }
	return ((i==nrows() && j==ncols()) ? 1 : status );
}

int RxMatdBase::print (FILE *fp, const char *s) const
{
	RxPrintSpec fmts;
	RxFmtNode *node;
	int nc, i, j, k;
	int numchars = 0;

	if (fmts.set(s) < 0)
	 { fprintf (fp, "RxMatdBase::print(): illegal format string\n");
	   return (-1);
	 }
	if (fmts.middle.head() == NULL)
	 { /* supply the default: %g */
	   fmts.middle.addTail (new RxFmtNode("%13.6e"));
	   fmts.suffix = "\n";
	 }
	if (fmts.prefix)
	 { if ((nc = fprintf (fp, fmts.prefix)) < 0)
	    { return (-1);
	    }
	   numchars += nc;
	 }
	if (fmts.middle.length() == 1)
	 { fmts.middle.addTail (new RxFmtNode(" ", fmts.middle.head()->fmt));
	 }
	/* print the first element using the first format. The remaining 
	   elements are printed by cycling through the remaining formats.
	   A new line is printed after each row, and the beginning of
	   each row is lined up with the column reached after printing
	   the prefix. */

	for (i=0; i<nrows(); i++)
	 { node = fmts.middle.head();
	   for (j=0; j<ncols(); j++)
	    { if ((nc = fprintf (fp, node->fmt, (*this)(i,j))) < 0)
	       { return (-1);
	       }
	      numchars += nc;
	      if ((node = node->next()) == NULL)
	       { node = fmts.middle.head()->next();
	       }
	    }
	   if (i<nrows()-1)
	    { putchar('\n');
	      numchars += 1;
	      for (k=0; k<fmts.prefixWidth; k++)
	       { putchar (' ');
		 numchars += 1;
	       }
	    }
	 }
	if (fmts.suffix)
	 { if ((nc = fprintf (fp, fmts.suffix)) < 0)
	    { return (-1);
	    }
	   numchars += nc;
	 }	
	return (numchars);
}

int RxMatdBase::print (char *buf, const char *s) const
{
	RxPrintSpec fmts;
	RxFmtNode *node;
	int nc, i, j, k;
	int numchars = 0;

	if (fmts.set(s) < 0)
	 { return (-1);
	 }
	if (fmts.middle.head() == NULL)
	 { /* supply the default: %g */
	   fmts.middle.addTail (new RxFmtNode("%13.6e"));
	   fmts.suffix = "\n";
	 }
	if (fmts.prefix)
	 { if ((nc = sprintf (buf+numchars, fmts.prefix)) < 0)
	    { return (-1);
	    }
	   numchars += nc;
	 }
	if (fmts.middle.length() == 1)
	 { fmts.middle.addTail (new RxFmtNode(" ", fmts.middle.head()->fmt));
	 }
	/* print the first element using the first format. The remaining 
	   elements are printed by cycling through the remaining formats.
	   A new line is printed after each row, and the beginning of
	   each row is lined up with the column reached after printing
	   the prefix. */

	for (i=0; i<nrows(); i++)
	 { node = fmts.middle.head();
	   for (j=0; j<ncols(); j++)
	    { if ((nc = sprintf (buf+numchars, node->fmt, (*this)(i,j))) < 0)
	       { return (-1);
	       }
	      numchars += nc;
	      if ((node = node->next()) == NULL)
	       { node = fmts.middle.head()->next();
	       }
	    }
	   if (i<nrows()-1)
	    { *(buf+numchars) = '\n';
	      numchars += 1;
	      for (k=0; k<fmts.prefixWidth; k++)
	       { *(buf+numchars) = ' ';
		 numchars += 1;
	       }
	    }
	 }
	if (fmts.suffix)
	 { if ((nc = sprintf (buf+numchars, fmts.suffix)) < 0)
	    { return (-1);
	    }
	   numchars += nc;
	 }	
	return (numchars);
}

#endif

#if 0
int RxMatdBase::scan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/0);
	return scan (&ci);
}

int RxMatdBase::xscan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/1);
	return scan (&ci);
}

int RxMatdBase::scan (const char *buf, int *cntp)
{
	RxCinput ci(buf, cntp);
	return scan (&ci);
}

istream &RxMatdBase::scan(istream &is)
{
	RxCinput ci(&is);
	int status;

	status = scan (&ci);
	if (status == EOF)
	 { is.clear (ios::eofbit);
	 }
	else if (status == 0)
	 { is.clear (ios::failbit);
	 }
	return is;
}

int RxMatdBase::scan (RxCinput *ci)
{
	RxParseObj *obj;

	obj = RxParseObj::scan (ci);
	if (obj == RxParseObj::eof)
	 { return EOF;
	 }
	else if (obj == NULL)
	 { return 0;
	 }
	else
	 { return oscan (obj);
	 }
}
#endif

int RxMatdBase::oscan (RxParseObj *obj)
{
	RxParseObj *op;	
	int i, j;

	if (!obj->isNumberList() || obj->getListLength() != nrows()*ncols())
	 { RxSysError (RxSysError::ParsingError,
"RxMatdBase::oscan: not a list of %d numbers\n", nrows()*ncols()).post();
	   return 0;
	 }
	op=obj->firstChild();
	for (i=0; i<nrows(); i++)
	 { for (j=0; j<ncols(); j++)
	    { (*this)(i,j) = op->getNumberValue();
	      op = obj->nextChild();
	    }
	 }
	// ci->setCnt (ci->nchars()); what was this???
	return 1;
}

#if 0
int RxMatdBase::print (FILE *fp, const char *s) const
{
	RxCoutput co(fp);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}

ostream &RxMatdBase::print(ostream &os) const
{
	RxCoutput co(&os);
	RxPrintSpec fs("%g");	// fake print spec -- format will be ignored
	cprint (&co, &fs);
	return os;
}

int RxMatdBase::print (char *buf, const char *s, int n) const
{
	RxCoutput co(buf, n);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}
#endif

int RxMatdBase::cprint (RxCoutput *co, RxPrintSpec *ps) const
{
	RxFmtNode *cf, *cf0, *cf1;
	const char *suffix;
	int i, j, k;

	co->output_str (ps->prefix());
	if (ps->numConvs() == 0) // then allocate our our format node
	 { cf0 = new RxFmtNode ("%g", 0, RxFmtNode::CONV);
	   suffix = "\n";
	 }
	else
	 { cf0 = (RxFmtNode *)ps->convNode(0);
	   suffix = ps->suffix();
	 }
	cf1 = cf0->next();
	if (cf1 == ps->suffixNode())
	 { cf1 = NULL;
	 }
	co->output_char ('[');
	for (i=0; i<nrows(); i++)
	 { cf = cf0;
	   for (j=0; j<ncols(); j++)
	    { co->output_double (cf->fmt(), (*this)(i,j));
	      if ((cf = cf->next()) == ps->suffixNode())
	       { cf = (cf1 ? cf1 : cf0);
	       }
	      if (j < ncols()-1)	// not last column
	       { if (cf && cf->type() == RxFmtNode::STR)
		  { co->output_str (cf->fmt());
		    cf = cf->next();
		  }
		 else
		  { co->output_str (" ");
		  }
	       }
	    }
	   if (i<nrows()-1)
	    { co->output_newline (ps->getLeadSpacing()+1);
	    }
	 }
	co->output_char (']');
	co->output_str (suffix);
	return co->nchars();
}

int RxMatdBase::equal (const RxMatdBase& m2, double eps) const
{
	double diff;
	int i, j;

	for (i=0; i<nrows(); i++)
	 { for (j=0; j<ncols(); j++)
	    { diff = m2(i,j) - (*this)(i,j);
	      if (ABS(diff) > eps)
	       { return (0);
	       }
	    }
	 }
	return (1);
}
