#include <RxMatdChol.h>
#include <RxMathError.h>
#include <stdlib.h>
#include <cdefs.h>
#include <math.h>
#include <mathx.h>

void RxMatdChol::doChol (const RxMatd &A)
{
	double tmp, anorm;
	int i, j, k;
	int w = L->width;
	double *buf = L->store;

	// Gaxpy Cholesky from Golub and Van Loan , "Matrix Computations"

	*L = A;

#define M(ii,jj)	buf[(ii)*w+(jj)]

	// get an approximate norm
	anorm = 0;
	for (i=0; i<n; i++)
	 { if (M(i,i) < 0)
	    { flags |= RX_CHOL_BADMATRIX;
	      return;
	    }
	   else if (M(i,i) > anorm)
	    { anorm = M(i,i);
	    }
	 }

	for (j=0; j<n; j++)
	 { 
	   if (j > 0)
	    { for (i=j; i<n; i++)
	       { tmp = 0;
		 for (k=0; k<j; k++)
		  { tmp += M(i,k)*M(j,k);
		  }
		 M(i,j) -= tmp;
	       }
	    }
	   tmp = M(j,j);
	   if (tmp < 0)
	    { flags |= RX_CHOL_BADMATRIX;
	    }
	   else
	    { tmp = sqrt(tmp);
	      if (anorm + tol*tmp == anorm)
	       { flags |= RX_CHOL_BADMATRIX;
	       }
	      else
	       { for (i=j; i<n; i++)
		  { M(i,j) /= tmp;
		  }
	       }
	    }
	 }
#undef M
}

int RxMatdChol::calc (const RxMatd &A)
{
	int newn;

	newn = A.nr;
	if (A.nc != newn)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdChol::calc: %K in A\n").raise();
	   flags &= ~RX_CHOL_CALCDONE;
	   return -1;
	 }
	n = newn;

	if (L)
	 { if (L->nr != n || L->nc != n)
	    { if (flags & RX_CHOL_L_MALLOCED)
	       { delete L;
		 L = new RxMatd (n, n);
	       }
	      else
	       { RxMathError (RxMathError::BadSize,
			      "RxMatdChol::calc: %K in L\n").raise();
		 flags &= ~RX_CHOL_CALCDONE;
		 return -1;
	       }
	    }
	 }
	else
	 { L = new RxMatd (n, n);
	   flags |= RX_CHOL_L_MALLOCED;
	 }
	doChol (A);
	flags |= RX_CHOL_CALCDONE;
	return (1);
}

int RxMatdChol::setL (RxMatd *Lm)
{
	if (L)
	 { if (flags & RX_CHOL_L_MALLOCED)
	    { delete L;
	      flags &= ~RX_CHOL_L_MALLOCED;
	    }
	   L = Lm;
	 }
	return (1);
}

void RxMatdChol::copyL (RxMatd *Lm) const
{
	int i, j;
	int w = L->width;

	if (!(flags & RX_CHOL_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdChol::copyL: %K\n").raise();
	   return;
	 }
	if ((Lm && (Lm->nr != n || Lm->nc != n)))
	 { RxMathError (RxMathError::BadSize, 
			"RxMatdChol::copyL: %K\n").raise();
	   return;
	 }
	if (Lm)
	 { for (i=0; i<n; i++)
	    { for (j=0; j<n; j++)
	       { Lm->store[Lm->width*i+j] = 
		    (i>=j) ? L->store[i*w+j] : 0.0;
	       }
	    }
	 }
}

RxMatdChol::~RxMatdChol ()
{
	if (flags & RX_CHOL_L_MALLOCED)
	 { delete L;
	 }
}

#define Lelem(ii,jj)	L->store[(ii)*w+(jj)]

RxVecd &RxMatdChol::solve (RxVecd *x, const RxVecd &b) const
{
	double sum;
	int w = L->width;
	int i, j;

	if (x->len != n || b.len != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdChol::solve: %K\n").raise();
	   return *x;
	 }
	if (!(flags & RX_CHOL_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdChol::solve: %K\n").raise();
	   return *x;
	 }

	for (i=0; i<n; i++)
	 { x->v[i] = b.v[i];
	 }
	for (i=0; i<n; i++)
	 { sum = x->v[i];
	   for (j=0; j<i; j++)
	    { sum -= x->v[j]*Lelem(i,j);
	    }
	   x->v[i] = sum/Lelem(i,i);
	 }
	for (i=n-1; i>=0; i--)
	 { sum = x->v[i];
	   for (j=i+1; j<n; j++)
	    { sum -= x->v[j]*Lelem(j,i);
	    }
	   x->v[i] = sum/Lelem(i,i);
	 }
	return *x;
}

#define MAXELEM	16

RxMatd &RxMatdChol::solve (RxMatd *X, const RxMatd &B) const
{
	double buf[MAXELEM];
	double *tmp;
	double sum;
	int w = L->width;
	int i, j, k;

	if (X->nr != n || B.nr != n || X->nc != B.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdChol::solve: %K\n").raise();
	   return *X;
	 }
	if (!(flags & RX_CHOL_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdChol::solve: %K\n").raise();
	   return *X;
	 }

	tmp = (n > MAXELEM) ? (double*)malloc(sizeof(double)*n) : buf;

	for (k=0; k<X->nc; k++)
	 { 
	   for (i=0; i<n; i++)
	    { tmp[i] = B.store[i*B.width+k];
	    }
	   for (i=0; i<n; i++)
	    { sum = tmp[i];
	      for (j=0; j<i; j++)
	       { sum -= tmp[j]*Lelem(i,j);
	       }
	      tmp[i] = sum/Lelem(i,i);
	    }
	   for (i=n-1; i>=0; i--)
	    { sum = tmp[i];
	      for (j=i+1; j<n; j++)
	       { sum -= tmp[j]*Lelem(j,i);
	       }
	      tmp[i] = sum/Lelem(i,i);
	      X->store[i*X->width+k] = tmp[i];
	    }
	 }
	if (n > MAXELEM) free (tmp);
	return (*X);
}

double RxMatdChol::condest(const RxMatd &A) const
{
	double pvbuf[MAXELEM];
	double pnbuf[MAXELEM];
	double ppbuf[MAXELEM];
	double yvbuf[MAXELEM];
	double *pvec;
	double *pneg;
	double *ppos;
	double *yvec;
	double ypos, yneg, yval;
	double pposNorm1;
	double pnegNorm1;
	double sum;
	int i, j, w;
	double zNormInf;
	double rNormInf;

	if (A.nr != n || A.nc != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdChol::condest: %K\n").raise();
	   return -1;
	 }
	if (!(flags & RX_CHOL_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdChol::condest: %K\n").raise();
	   return -1;
	 }

	w = L->width;

	if (n > MAXELEM)
	 { pvec = (double*)malloc(n*3*sizeof(double));
	   pneg = pvec + n;
	   ppos = pneg + n;
	   yvec = ppos + n;
	 }
	else
	 { pvec = pvbuf;
	   pneg = pnbuf;
	   ppos = ppbuf;
	   yvec = yvbuf;
	 }

	for (j=0; j<n; j++)
	 { pvec[j] = 0;
	 }
	for (j=0; j<n; j++)
	 { pposNorm1 = 0;
	   pnegNorm1 = 0;
	   ypos = ( 1 - pvec[j])/Lelem(j,j);
	   for (i=j+1; i<n; i++)
	    { ppos[i] = pvec[i] + ypos*Lelem(i,j);
	      pposNorm1 += ABS(ppos[i]);
	    }
	   yneg = (-1 - pvec[j])/Lelem(j,j);
	   for (i=j+1; i<n; i++)
	    { pneg[i] = pvec[i] + yneg*Lelem(i,j);
	      pnegNorm1 += ABS(pneg[i]);
	    }
	   if (ABS(ypos) + pposNorm1 >= ABS(yneg) + pnegNorm1)
	    { yvec[j] = ypos;
	      for (i=j+1; i<n; i++)
	       { pvec[i] = ppos[i];
	       }
	    }
	   else
	    { yvec[j] = yneg;
	      for (i=j+1; i<n; i++)
	       { pvec[i] = pneg[i];
	       }
	    }
	 }

	rNormInf = 0;
	zNormInf = 0;

	// Solve L^T r = y

	for (i=n-1; i>=0; i--)
	 { sum = 0;
	   for (j=i+1; j<n; j++)
	    { sum += yvec[j]*Lelem(j,i);
	    }
	   yval = (yvec[i]-sum)/Lelem(i,i);
	   if (ABS(yval) > rNormInf)
	    { rNormInf = ABS(yval);
	    }
	   yvec[i] = yval;
	 }

	// Solve L w = P r

	for (i=0; i<n; i++)
	 { sum = yvec[i];
	   for (j=0; j<i; j++)
	    { sum -= yvec[j]*Lelem(i,j);
	    }
	   yvec[i] = sum/Lelem(i,i);
	 }

	// Solve U z = w

	for (i=n-1; i>=0; i--)
	 { sum = 0;
	   for (j=i+1; j<n; j++)
	    { sum += yvec[j]*Lelem(j,i);
	    }
	   yval = (yvec[i]-sum)/Lelem(i,i);
	   if (ABS(yval) > zNormInf)
	    { zNormInf = ABS(yval);
	    }
	   yvec[i] = yval;
	 }

	// Compute the infinity norm of A

	if (n > MAXELEM)
	 { free (pvec);
	 }

	return (A.normInf() * zNormInf / rNormInf);
}

double RxMatdChol::det() const
{
	double prod;
	int w = L->width;
	int i;

	if (!(flags & RX_CHOL_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdChol::det: %K\n").raise();
	   return -1;
	 }
	if (flags & RX_CHOL_BADMATRIX)
	 { return (0.0);
	 }
	prod = 1.0;
	for (i=0; i<n; i++)
	 { prod *= Lelem(i,i);
	 }
	return (prod*prod);
}

RxMatd &RxMatdChol::inv (RxMatd *M) const
{
	if (M->nc != n || M->nr != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdChol::inv: %K\n").raise();
	   return *M;
	 }
	if (!(flags & RX_CHOL_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdChol::inv: %K\n").raise();
	   return *M;
	 }
	M->setIdent();
	solve (M, *M);	
	return *M;
}
