#include <RxMatdSvd.h>
#include <RxMathError.h>
#include <stdlib.h>
#include <cdefs.h>
#include <math.h>
#include <mathx.h>

RxMatd RxMatdSvd::none(0,0,(double*)NULL);

void house (double *v, double *x, int n, int w)
{
	double xi, mu, beta;
	int i;

	mu = 0;
	for (i=0; i<n; i++)
	 { xi = *x;
	   mu += xi*xi;
	   v[i] = xi;
	   x += w;
	 }
	mu = sqrt(mu);
	if (mu != 0)
	 { beta = v[0] + SGN(v[0])*mu;
	   for (i=1; i<n; i++)
	    { v[i] /= beta;
	    }
	 }
	v[0] = 1;
}

#define M(ii,jj)	buf[(ii)*w+(jj)]
#define MAXM	256
#define MAX_ITER 100

#if 0
static void printMat (char *s, double *buf, int m, int n, int w)
{
	int i, j;

	if (s)
	 { printf ("%s\n", s);
	 }
	for (i=0; i<m; i++)
	 { for (j=0; j<n; j++)
	    { printf ("%13.6e ", M(i,j));
	    }
	   printf ("\n");
	 }
}
#endif

static void houseRow (double *buf, double *v, int m, int n, int w)
{
	/* Recombine the rows 0..m of buf according to the 
	   householder vector v */
	   
	double d, beta, wj;
	int i, j;

	d = 0;
	for (i=0; i<m; i++)
	 { d += v[i]*v[i];
	 }
	beta = -2/d;
	for (j=0; j<n; j++)
	 { wj = 0;
	   for (i=0; i<m; i++)
	    { wj += M(i,j)*v[i];
	    }
	   for (i=0; i<m; i++)
	    { M(i,j) += beta*v[i]*wj;
	    }
	 }
}

static void houseCol (double *buf, double *v, int m, int n, int w)
{
	/* Recombine the columns 0..n of buf according to the 
	   householder vector v */

	double d, beta, wi;
	int i, j;

	d = 0;
	for (j=0; j<n; j++)
	 { d += v[j]*v[j];
	 }
	beta = -2/d;
	for (i=0; i<m; i++)
	 { wi = 0;
	   for (j=0; j<n; j++)
	    { wi += M(i,j)*v[j];
	    }
	   for (j=0; j<n; j++)
	    { M(i,j) += beta*wi*v[j];
	    }
	 }
}

static void houseRowAccum (double *Q, double *buf, int m, int n, int Qw)
{
	// given a set of householder row transforms Pn-1 Pn-2 ... P0 = Q
	// stored implictly in buf, this routine computes Q^T

	double v[MAXM];
	int j, k;
	int w = n;

	for (j=n-1; j>=0; j--)
	 { v[0] = 1;
	   for (k=1; k<m-j; k++)
	    { v[k] = M(j+k,j);
	    }
	   houseRow (&Q[j*Qw+j], v, m-j, n-j, Qw);
	 }	
}

static void houseColAccum (double *Q, double *buf, int n, int Qw)
{
	// given a set of householder column transforms P0 P1 ... Pn-1 = Q
	// stored implictly in buf, this routine computes Q^T

	double v[MAXM];
	int j, k;
	int w = n;

	for (j=n-3; j>=0; j--)
	 { v[0] = 1;
	   for (k=j+2; k<n; k++)
	    { v[k-j-1] = M(j,k);
	    }
	   houseCol (&Q[(j+1)*Qw+j+1], v, n-j-1, n-j-1, Qw);
	 }	
}

static void bidiagonalize (double *buf, int m, int n)
{
	double v[MAXM];
	int i, j, k;
	int w = n;

	for (j=0; j<n; j++)
	 { house (v, &M(j,j), m-j, n);
	   houseRow (&M(j,j), v, m-j, n-j, n);
	   for (i=j+1; i<m; i++)
	    { M(i,j) = v[i-j];
	    }
	   if (j < n-2)
	    { house (v, &M(j,j+1), n-j-1, 1);
	      houseCol (&M(j,j+1), v, m-j, n-j-1, n);
	      for (k=j+2; k<n; k++)
	       { M(j,k) = v[k-j-1];
	       }
	    }
	 }
}

#define C	sincos[0]
#define S	sincos[1]

static void givens (double *sincos, double a, double b)
{
	double tau;

	if (b == 0)
	 { C = 1;
	   S = 0;
	 }
	else
	 { if (ABS(b) > ABS(a))
	    { tau = -a/b;
	      S = 1/sqrt(1+tau*tau);
	      C = S*tau;
	    }
	   else
	    { tau = -b/a;
	      C = 1/sqrt(1+tau*tau);
	      S = C*tau;
	    }
	 }
}

#if 0
static void givensRow (double *buf, int n, int w, double *sincos, int k,int l)
{
	// pre-multiples A by a sequence of Givens rotations given by
	// 
	//  [  c   -s  ]
	//  [          ]
	//  [  s    c  ]
	//
	double ak, al;
	int j;

	for (j=0; j<n; j++)
	 { ak = M(k, j);
	   al = M(l, j);
	   M(k, j) = ak*C - al*S;
	   M(l, j) = ak*S + al*C;
	 }
}

static void givensRowVec (double *vr, double *v1, double *sincos, int k, int l)
{
	// pre-multiples v by a sequence of Givens rotations given by
	//
	//  [  c   -s  ]
	//  [          ]
	//  [  s    c  ]
	//
	double ak, al;
 
	ak = v1[k];
	al = v1[l];
	vr[k] = ak*C - al*S;
	vr[l] = ak*S + al*C;
}

static void matVecMult (double *vr, double *A, int m, int n, double *v1)
{
	double tmp;
	int i, j;

	for (i=0; i<m; i++)
	 { tmp = 0;
	   for (j=0; j<n; j++)
	    { tmp += A[i*n+j]*v1[j];
	    }
	   vr[i] = tmp;
	 }
}
#endif

static void givensCol (double *buf, int m, int w, double *sincos, int k,int l)
{
	// post-multiples A by a Givens rotation
	// 
	//  [  c    s  ]
	//  [          ]
	//  [ -s    c  ]
	//

	double ak, al;
	int i;

	for (i=0; i<m; i++)
	 { ak = M(i, k);
	   al = M(i, l);
	   M(i, k) = ak*C - al*S;
	   M(i, l) = ak*S + al*C;
	 }
}

void RxMatdSvd::takeStep (double *buf, int ib, int nn, int w)
{
	double bmm, bmn, bnn, blm;
	double tmm, tmn, tnn;
	double b11, b12;
	double sincos[2];
	double d, mu;
	double y, z;
	int k;

	blm = (nn >= 3 ? M(nn-3,nn-2) : 0);
	bmm = M(nn-2,nn-2);
	bmn = M(nn-2,nn-1);
	bnn = M(nn-1,nn-1);

	tmm = blm*blm + bmm*bmm;
	tmn = bmm*bmn;
	tnn = bmn*bmn + bnn*bnn;

	b11 = M(0,0);
	b12 = M(0,1);

	d = (tmm-tnn)/2;
	mu = tnn - tmn*tmn/(d + SGN(d)*sqrt(d*d + tmn*tmn));

	y = b11*b11-mu;
	z = b11*b12;

	// At every step of the iteration, we are concerned with the 
	// following elements:
	//
	//	bk01  bk02            M(k-1,k) M(k-1,k+1)
	//	bk11  bk12  bk13  ==  M(k,k)   M(k,k+1)   M(k,k+2)
	//	bk21  bk22  bk23      M(k+1,k) M(k+1,k+1) M(k+1,k+2)
	// 
	// The situtation at the beginning of the loop, with k=0,
	// followed by the first and second givens multiplies, looks like:
	//
	//   -  -		-  -		-  -
	//   X  X  0		X  X  0		X  X  +
	//   0  X  X		+  X  X		0  X  X
	//
	// The situtation for subsequent k>0 looks like:
	//
	//   X  +		X  0		X  0
	//   X  X  0		X  X  0		X  X  +
	//   0  X  X		+  X  X		0  X  X
	// 
	// and the situation for k == nn-2 looks like
	//	   
	//   X  +		X  0		X  0
	//   X  X  -		X  X  -		X  X  -
	//   0  X  -		+  X  -		0  X  -
	
	double bk01, bk02;
	double bk11, bk12, bk13;
	double bk21, bk22, bk23;
	double btmp;

	bk01 = bk02 = bk21 = 0;
	bk11 = M(0,0);
	bk12 = M(0,1);
	bk22 = M(1,1);
	bk23 = M(1,2);

	for (k=0; k<nn-1; k++)
	 { 
	   if (k==0)
	    { givens (sincos, b11*b11-mu, b11*b12);
//printf ("mu=%18.15e  %18.15e %18.15e\n", mu, b11*b11-mu, b11*b12);
	    }
	   else
	    { givens (sincos, bk01, bk02);
	    }
//printf ("%d A: C=%18.15e S=%18.15e\n", k, C, S);
//printf ("bk11=%13.8e bk12=%13.8e\n", bk11, bk12);
	   bk01 = C*bk01 - S*bk02;
	   btmp = C*bk11 - S*bk12;
//printf ("btmp=%13.8e\n", btmp);
	   bk21 =        - S*bk22;
	   bk12 = S*bk11 + C*bk12;
	   bk22 =          C*bk22;
	   bk11 = btmp;
	   if (V != RX_SVD_OMIT)
	    { givensCol (V->store, n, n, sincos, ib+k, ib+k+1);
	    }

	   if (k != 0)
	    { M(k-1,k) = bk01;
	    }
	   givens (sincos, bk11, bk21);
//printf ("%d B: C=%18.15e S=%18.15e\n", k, C, S);
//printf ("bk11=%13.8e bk21=%13.8e\n", bk11, bk21);
	   bk11 = C*bk11 - S*bk21;
	   btmp = C*bk12 - S*bk22;
	   bk13 =        - S*bk23;
	   bk22 = S*bk12 + C*bk22;
	   bk23 =        + C*bk23;
	   bk12 = btmp;
	   if (U != RX_SVD_OMIT)
	    { givensCol (U->store, m, n, sincos, ib+k, ib+k+1);
	    }

	   M(k,k) = bk11;
	   M(k,k+1) = bk12;
	   M(k+1,k+1) = bk22;

	   if (k == nn-2)
	    { M(k,k+1) = bk12;
	      M(k+1,k+1) = bk22;
	    }
	   else 
	    { bk01 = bk12;  bk02 = bk13;
	      bk11 = bk22;  bk12 = bk23;
	      bk22 = M(k+2,k+2);
	      if (k < nn-2)
	       { bk23 = M(k+2,k+3);
	       }
	      else
	       { bk23 = 0;
	       }
	    }
	 }
}

void RxMatdSvd::zeroRow (double *buf, int ib, int p, int w)
{
	//	A has the following form,
	//	
	//	 |< p >|< q >|
	//	0 X
	//	  X X
	//	    X X
	//	      X
	//	        X
	//	          X
	//                  X
	//
	//	and so we want to pre-multiply it by a series
	//	of givens rotations so as to zero the top row.
	//	It is assumed that p is non-zero.
	//
	//	At each step, the critical elements of the
	//	iteration are identified as follows:
	//
	//	+  0		al1  al2
	//	.  .	==	
	//	X  X		ak1  ak2
	//

	double al1, ak1, ak2;
	double al1next;
	double sincos[2];
	int k;

	al1next = M(0,1);
	for (k=1; k<=p; k++)
	 { al1 = al1next;
	   ak1 = M(k,k);
	   givens (sincos, ak1, al1);
	   if (U != RX_SVD_OMIT)
	    { givensCol (U->store, m, n, sincos, k+ib, ib);
	    }
	   M(k,k) = ak1*C - al1*S;
	   if (k<p)
	    { ak2 = M(k,k+1);
	      M(k,k+1) = ak2*C;
	      al1next = ak2*S;
	    }
	 }
	M(0,1) = 0;
}

#undef C
#undef S

void RxMatdSvd::swapUV ()
{
	RxMatd *T;
	int d;

	d = n;
	n = m;
	m = d;

	T = U;
	U = V;
	V = T;
}

int RxMatdSvd::doSvd ()
{
	double anorm = 0; // lower bound approximation of the 2 norm
	int UVswapped = 0;
	int i, j, w;
	int kd;	// starting index of lower diagonal
	int kb;	// starting index of lowest full-bidiagonal
	int zerod;
	int icnt = 0;

	if (m < n)
	 { swapUV();
	   UVswapped = 1;
	 }
	w = n;

	flags &= ~RX_SVD_CALCDONE;

//printMat ("doSvd", buf, m, n, w);

	if (U != RX_SVD_OMIT)
	 { U->setIdent();
	 }
	if (V != RX_SVD_OMIT)
	 { V->setIdent();
	 }
	bidiagonalize (buf, m, n);
	if (U != RX_SVD_OMIT)
	 { houseRowAccum (U->store, buf, m, n, U->width);
	 }
	if (V != RX_SVD_OMIT)
	 { houseColAccum (V->store, buf, n, V->width);
	   V->tsp();
	 }
	// estimate matrix norm 
	for (i=0; i<n-1; i++)
	 { anorm = MAX(anorm, ABS(M(i,i)));
	   anorm = MAX(anorm, ABS(M(i,i+1)));
	 }
	anorm = MAX(anorm, ABS(M(n-1,n-1)));
	
#if 1
	kd = n;
	while (kd != 0)
	 { 
	   for (i=0; i<kd-1; i++)
	    { if (ABS(M(i,i+1)) < tol*(ABS(M(i,i)) + ABS(M(i+1,i+1))))
	       { M(i,i+1) = 0;
	       }
	    }
	   while (kd > 0)
	    { if (kd > 1 && M(kd-2,kd-1) != 0)
	       { break;
	       }
	      kd--;
	    }
//printf ("kd=%d\n", kd);
//printMat ("iter", buf, m, n, w);
	   if (kd > 0)
	    { kb = kd-2;
	      zerod = 0;
	      while (kb > 0 && M(kb,kb) != 0 && M(kb-1,kb) != 0)
	       { kb--;
	       }
	      if (anorm + M(kb,kb) == anorm)
	       { M(kb,kb) = 0;
		 zeroRow (&M(kb,kb), kb, kd-kb-1, n);
//printf ("zeroRow\n");
	       }
	      else
	       { 
//printMat ("before takeStep", buf, m, n, w);
		 takeStep (&M(kb,kb), kb, kd-kb, n);
//printMat ("after takeStep", buf, m, n, w);
	       }
	    }
	   if (++icnt >= MAX_ITER)
	    { fprintf (stderr,
"RxMatdSvd::doSvd(): failed to converge after %d iterations\n", icnt);
//printf ("tol=%10.6e kd=%d\n", tol, kd);
//printMat ("end", buf, m, n, w);
	      exit (1);
	    }
	 }
//printMat ("ok  ",  buf, m, n, w);
#endif
	 { double s;

	   sigmin = 0;
	   sigmax = HUGE_VAL;
	   sigdet = 1;
	   for (j=0; j<n; j++)
	    { s = buf[j*n+j];
	      sigdet *= s;
	      if (s < 0)
	       { s = -s;
		 if (V != RX_SVD_OMIT)
		  { for (i=0; i<n; i++)
		     { V->store[i*V->width+j] = -V->store[i*V->width+j];
		     }
		  }
	       }
	      if (s > sigmax)
	       { sigmax = s;
	       }
	      if (s < sigmin)
	       { sigmin = s;
	       }
	      sig->v[j] = s;
	    }
	   if (n == 1)
	    { sigdet = -sigdet;
	    }
	 }
	if (UVswapped)
	 { swapUV();
	 }
	flags |= RX_SVD_CALCDONE;
	return (1);
}

#undef M

int RxMatdSvd::calcSetup (const RxMatd &A, RxVecd *r, const RxVecd *b)
{
	int i, j;
	int mind;

	if (r)
	 { rv = r;
	   bv = b;
	 }
	else
	 { rv = NULL;
	   bv = NULL;
	 }
	if (buf && m*n != A.nr*A.nc)
	 { free (buf);
	   buf = NULL;
	 }
	m = A.nr;
	n = A.nc;
	mind = MIN(m,n);

	if (buf == NULL)
	 { buf = (double*)malloc (m*n*sizeof(double));
	 }

	if (sig)
	 { if (sig->len != mind)
	    { if (flags & RX_SVD_SIG_MALLOCED)
	       { delete sig;
		 sig = new RxVecd (mind);
	       }
	      else
	       { RxMathError (RxMathError::BadSize,
			      "RxMatdSvd::calc: %K in sig\n").raise();
		 return -1;
	       }
	    }
	 }
	else
	 { sig = new RxVecd (mind);
	   flags |= RX_SVD_SIG_MALLOCED;
	 }
	if (U)
	 { if (U->nr != m || U->nc != mind)
	    { if (flags & RX_SVD_U_MALLOCED)
	       { delete U;
		 U = new RxMatd (m,mind);
	       }
	      else
	       { RxMathError (RxMathError::BadSize,
			      "RxMatdSvd::calc: %K in U\n").raise();
		 return -1;
	       }
	    }
	 }
	else
	 { U = new RxMatd (m,mind);
	   flags |= RX_SVD_U_MALLOCED;
	 }
	if (V)
	 { if (V->nc != mind || V->nr != n)
	    { if (flags & RX_SVD_V_MALLOCED)
	       { delete V;
		 V = new RxMatd (n,mind);
	       }
	      else
	       { RxMathError (RxMathError::BadSize,
			      "RxMatdSvd::calc: %K in V\n").raise();
		 return -1;
	       }
	    }
	 }
	else
	 { V = new RxMatd (n,mind);
	   flags |= RX_SVD_V_MALLOCED;
	 }
	if (rv && (rv->len != m || bv->len != m))
	 { RxMathError (RxMathError::BadSize,
			      "RxMatdSvd::calc: %K in r or b\n").raise();
	   return -1;
	 }
	if (m >= n)
	 { for (i=0; i<m; i++)
	    { for (j=0; j<n; j++)
	       { buf[i*n+j] = A.store[i*A.width+j];
	       }
	    }
	 }
	else
	 { for (i=0; i<m; i++)
	    { for (j=0; j<n; j++)
	       { buf[j*m+i] = A.store[i*A.width+j];
	       }
	    }
	 }
	return 1;
}

int RxMatdSvd::calc (const RxMatd &A)
{
	if (calcSetup (A, NULL, NULL) <0)
	 { return -1;
	 }
	return (doSvd ());
}

int RxMatdSvd::calc (const RxMatd &A, RxVecd *r, const RxVecd &b)
{
	if (calcSetup (A, r, &b) <0)
	 { return -1;
	 }
	return (doSvd ());
}

int RxMatdSvd::setSUV (RxVecd *s, RxMatd *u, RxMatd *v)
{
	if (s)
	 { if (flags & RX_SVD_SIG_MALLOCED)
	    { delete s;
	      flags &= ~RX_SVD_SIG_MALLOCED;
	    }
	   sig = s;
	 }
	if (u)
	 { if (flags & RX_SVD_U_MALLOCED)
	    { delete U;
	      flags &= ~RX_SVD_U_MALLOCED;
	    }
	   U = u;
	 }
	if (v)
	 { if (flags & RX_SVD_V_MALLOCED)
	    { delete V;
	      flags &= ~RX_SVD_V_MALLOCED;
	    }
	   V = v;
	 }
	return (1);
}

void RxMatdSvd::getSUV (RxVecd **s, RxMatd **u, RxMatd **v)
{
	*s = sig;
	*u = U;
	*v = V;
}

RxMatdSvd::~RxMatdSvd ()
{
	if (flags & RX_SVD_SIG_MALLOCED)
	 { delete sig;
	 }
	if (flags & RX_SVD_U_MALLOCED)
	 { delete U;
	 }
	if (flags & RX_SVD_V_MALLOCED)
	 { delete V;
	 }
}

RxVecd &RxMatdSvd::solve (RxVecd *x, const RxVecd &b) const
{
	int mind = MIN(m,n);
	int i;

	if (x->len != n || b.len != m)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdSvd::solve: %K\n").raise();
	   return *x;
	 }
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::solve: %K\n").raise();
	   return *x;
	 }
	RxVecd vtmp(mind);
	U->Vmul (&vtmp, b);
	for (i=0; i<mind; i++)
	 { vtmp.v[i] /= sig->v[i];
	 }
	V->mulV (x, vtmp);
	return (*x);
}

RxVecd &RxMatdSvd::Tsolve (RxVecd *x, const RxVecd &b) const
{
	int mind = MIN(m,n);
	int i;

	if (x->len != m || b.len != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdSvd::Tsolve: %K\n").raise();
	   return *x;
	 }
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::solve: %K\n").raise();
	   return *x;
	 }
	RxVecd vtmp(mind);
	V->Vmul (&vtmp, b);
	for (i=0; i<mind; i++)
	 { vtmp.v[i] /= sig->v[i];
	 }
	U->mulV (x, vtmp);
	return (*x);
}

#define MAXELEM	16

RxMatd &RxMatdSvd::solve (RxMatd *X, const RxMatd &B) const
{
	int mind = MIN(m,n);
	int i, j;

	if (X->nr != n || B.nr != m || X->nc != B.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdSvd::solve: %K\n").raise();
	   return *X;
	 }
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::solve: %K\n").raise();
	   return *X;
	 }
	RxMatd Mtmp(mind, B.nc);
	Mtmp.Tmul (*U, B);
Mtmp.print ("tmp = [%9.5f]\n");
	for (j=0; j<B.nc; j++)
	 { for (i=0; i<mind; i++)
	    { Mtmp.store[i*Mtmp.width+j] /= sig->v[i];
	    }
	 }
Mtmp.print ("div = [%9.5f]\n");
	X->mul (*V, Mtmp);
X->print ("X   = [%9.5f]\n");
	return (*X);
}

RxMatd &RxMatdSvd::Tsolve (RxMatd *X, const RxMatd &B) const
{
	int mind = MIN(m,n);
	int i, j;

	if (X->nr != m || B.nr != n || X->nc != B.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdSvd::solve: %K\n").raise();
	   return *X;
	 }
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::solve: %K\n").raise();
	   return *X;
	 }
	RxMatd Mtmp(mind, B.nc);
	Mtmp.Tmul (*V, B);
	for (j=0; j<B.nc; j++)
	 { for (i=0; i<mind; i++)
	    { Mtmp.store[i*Mtmp.width+j] /= sig->v[i];
	    }
	 }
	X->mul (*U, Mtmp);
	return (*X);
}

double RxMatdSvd::cond() const
{
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::cond: %K\n").raise();
	   return -1;
	 }
	else
	 { return (sigmax/sigmin);
	 }
}

double RxMatdSvd::norm2() const
{
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::norm2: %K\n").raise();
	   return -1;
	 }
	else
	 { return (sigmax);
	 }
}

double RxMatdSvd::det() const
{
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::det: %K\n").raise();
	   return -1;
	 }
	else
	 { return (sigdet);
	 }
}

RxMatd &RxMatdSvd::inv (RxMatd *M) const
{
	if (M->nc != n || M->nr != m)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdSvd::inv: %K\n").raise();
	   return *M;
	 }
	if (!(flags & RX_SVD_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdSvd::inv: %K\n").raise();
	   return *M;
	 }
	else
	 { M->setIdent();
	   solve (M, *M);	
	   return *M;
	 }
}
