#include <RxGenKin.h>
#include <RxVecd6.h>
#include <RxGenError.h>
#include <RxSysUtil.h>
#include <RxSysError.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cdefs.h>
#include <stdiox.h>
#include <ctype.h>

void RxGenKin::init ()
{
	dhpSet = NULL;
	jacob = 0;
	lambda = 0;
	flags = 0;
	numj = 0;
	tol = RX_GENKIN_DEFAULT_TOL;
	lu = NULL;
	JJ = NULL;
	svd = NULL;
	condlimit = (1e6);
}

int RxGenKin::set (const char *name)
{
	DHParams *dhp;
	FILE *fp;
	char *fileName = (char*)malloc(strlen(name)+5);
	char *pathName;
	int j, n;
	
	strcpy (fileName, name);
	strcat (fileName, ".kin");

	numj = -1;
	if (dhpSet)
	 { delete dhpSet;
	   dhpSet = NULL;
	 }

	fp = locateFile (&pathName, "LIMA_ROBOT_PATH", fileName);
	if (fp == NULL)
	 { return (-1);
	 }
	xscanfBegin (fp);
	if (xscanf (fp, "%d", &n) != 1 || n <= 0)
	 { RxSysError (RxSysError::ParsingError,
"RxGenKin::set %K, line %d, in %s\n", xscanfLine(fp), pathName);
	   return (-1);
	 }

	dhpSet = new DHParams[n];
	for (j=0; j<n; j++)
	 { dhp = &dhpSet[j];
	   if (xscanf (fp, "%c", &dhp->type) != 1 ||
	       strchr ("rRpP", dhp->type) == NULL ||
	       xscanf (fp, "%lf", &dhp->a) != 1 ||
	       xscanf (fp, "%lf", &dhp->d) != 1 ||
	       xscanf (fp, "%lf", &dhp->alpha) != 1 ||
	       xscanf (fp, "%lf", &dhp->theta) != 1)
	    { RxSysError (RxSysError::ParsingError,
"RxGenKin::set %K, line %d, in %s\n", xscanfLine(fp), pathName);
	      delete dhpSet;
	      dhpSet = NULL;
	      return (-1);
	    }
	   if (islower(dhp->type))
	    { dhp->type = toupper(dhp->type);
	    }
	   dhp->theta *= DTOR;
	   dhp->alpha *= DTOR;
	   dhp->sA = sin(dhp->alpha);
	   dhp->cA = cos(dhp->alpha);
	   dhp->sT = sin(dhp->theta);
	   dhp->cT = cos(dhp->theta);
	 }
	numj = n;
	return (1);
}

int RxGenKin::set (const DHParams *params, int n)
{
	DHParams *dhp;
	int j;

	numj = -1;
	if (dhpSet)
	 { delete dhpSet;
	   dhpSet = NULL;
	 }

	if (n <= 0)
	 { return (-1);
	 }
	for (j=0; j<n; j++)
	 { if (strchr ("rRpP", params[j].type) == NULL)
	    { return (-1);
	    }
	 }
	dhpSet = new DHParams [n];

	for (j=0; j<n; j++)
	 { dhp = &dhpSet[j];
	   dhp->type = toupper(params[j].type);
	   dhp->a = params[j].a;
	   dhp->d = params[j].d;
	   dhp->alpha = params[j].alpha;
	   dhp->theta = params[j].theta;

	   dhp->sA = sin(dhp->alpha);
	   dhp->cA = cos(dhp->alpha);
	   dhp->sT = sin(dhp->theta);
	   dhp->cT = cos(dhp->theta);
	 }
	numj = n;
	return (1);
}

int RxGenKin::set (const RxGenKin &kin)
{
	int j;

	numj = -1;
	if (dhpSet)
	 { delete dhpSet;
	   dhpSet = NULL;
	 }
	if (kin.numj == -1)
	 { return (-1);
	 }
	dhpSet = new DHParams[kin.numj];
	for (j=0; j<kin.numj; j++)
	 { dhpSet[j] = kin.dhpSet[j];
	 }
	numj = kin.numj;
	return (1);
}

int RxGenKin::get (DHParams *params)
{
	int j;

	for (j=0; j<numj; j++)
	 { params[j] = dhpSet[j];
	 }
	return (numj);
}

#define CA	dhp->cA
#define SA	dhp->sA
#define CT	dhp->cT
#define ST	dhp->sT

int RxGenKin::setJnts (const RxVecd &jnts)
{
	DHParams *dhp;
	int j;

	if (numj != jnts.length())
	 { RxGenError (RxGenError::BadSize,
		       "RxGenKin::setJnts: %k\n").raise();
	   return -1;
	 }
	dhp = dhpSet;
	for (j=0; j<numj; j++)
	 { if (dhp->type == 'P')
	    { dhp->d = jnts(j);
	    }
	   else
	    { dhp->theta = jnts(j);
	      ST = sin(dhp->theta);
	      CT = cos(dhp->theta);
	    }
	   dhp++;
	 }
	flags = 0;
	return 1;
}

int RxGenKin::getJnts (RxVecd *jnts)
{
	DHParams *dhp;
	int j;

	if (numj != jnts->length())
	 { RxGenError (RxGenError::BadSize,
		       "RxGenKin::setJnts: %k\n").raise();
	   return -1;
	 }
	dhp = dhpSet;
	for (j=0; j<numj; j++)
	 { if (dhp->type == 'P')
	    { (*jnts)(j) = dhp->d;
	    }
	   else
	    { (*jnts)(j) = dhp->theta;
	    }
	   dhp++;
	 }
	return 1;
}

RxGenKin::~RxGenKin ()
{
	if (jacob) delete jacob;
	if (dhpSet) delete dhpSet;
	if (lu) delete lu;
	if (svd) delete svd;
	if (JJ) delete JJ;
}

int RxGenKin::fwdK (RxTrsf *Tr, int *conf)
{
	RxTrsf Am;
	int j;

	if (conf)
	 { *conf = 0;
	 }
	A (Tr, 0);
	for (j=1; j<numj; j++)
	 { A (&Am, j);
	   Tr->mul (Am);
	 }
	return (0);
}

RxTrsf& RxGenKin::A (RxTrsf *Ar, int j, double jval)
{
	DHParams *dhp;

	// General form of an A matrix:
	//
	// [ CT	   -ST CA     ST SA	a CT ]
	// [ ST     CT CA    -CT SA     a ST ]
	// [  0      SA        CA        d   ]

	dhp = &dhpSet[j];
	if (dhp->type == 'P')
	 { dhp->d = jval;
	 }
	else
	 { dhp->theta = jval;
	   ST = sin(jval);
	   CT = cos(jval);
	 }
	Ar->R.n.x() = CT;
	Ar->R.n.y() = ST;
	Ar->R.n.z() = 0;

	Ar->R.o.x() = -ST*CA;
	Ar->R.o.y() = CT*CA;
	Ar->R.o.z() = SA;

	Ar->R.a.x() = ST*SA;
	Ar->R.a.y() = -CT*SA;
	Ar->R.a.z() = CA;

	Ar->p.x() = dhp->a*CT;
	Ar->p.y() = dhp->a*ST;
	Ar->p.z() = dhp->d;

	flags = 0;
	return (*Ar);
}

RxTrsf& RxGenKin::A (RxTrsf *Ar, int j)
{
	DHParams *dhp;

	dhp = &dhpSet[j];
	Ar->R.n.x() = CT;
	Ar->R.n.y() = ST;
	Ar->R.n.z() = 0;

	Ar->R.o.x() = -ST*CA;
	Ar->R.o.y() = CT*CA;
	Ar->R.o.z() = SA;

	Ar->R.a.x() = ST*SA;
	Ar->R.a.y() = -CT*SA;
	Ar->R.a.z() = CA;

	Ar->p.x() = dhp->a*CT;
	Ar->p.y() = dhp->a*ST;
	Ar->p.z() = dhp->d;
	return (*Ar);
}

void RxGenKin::buildJacobian()
{
	RxTrsf Um, Am;
	int j;

	if (jacob == NULL)
	 { jacob = new RxMatd (6, numj);
	 }
	for (j=numj-1; j>=0; j--)
	 { 
	   if (j==numj-1)
	    { A (&Um, j);
	    }
	   else
	    { A (&Am, j);
	      Um.mul (Am, Um);
	    }
	   if (dhpSet[j].type == 'P')
	    { (*jacob)(0,j) = Um.R.n.z();
	      (*jacob)(1,j) = Um.R.o.z();
	      (*jacob)(2,j) = Um.R.a.z();

	      (*jacob)(3,j) = 0;
	      (*jacob)(4,j) = 0;
	      (*jacob)(5,j) = 0;
	    }
	   else
	    { (*jacob)(0,j) = Um.p.x()*Um.R.n.y() - Um.p.y()*Um.R.n.x();
	      (*jacob)(1,j) = Um.p.x()*Um.R.o.y() - Um.p.y()*Um.R.o.x();
	      (*jacob)(2,j) = Um.p.x()*Um.R.a.y() - Um.p.y()*Um.R.a.x();

	      (*jacob)(3,j) = Um.R.n.z();
	      (*jacob)(4,j) = Um.R.o.z();
	      (*jacob)(5,j) = Um.R.a.z();
	    }
	 }
	flags |= RX_GENKIN_JACOBIAN_VALID;
}

void RxGenKin::invertJacobian()
{
	if (lu == NULL)
	 { lu = new RxMatdLu ();
	 }
	if (numj != 6)
	 { JJ = new RxMatd (MIN(numj,6), MIN(numj,6));
	 }

	if (numj == 6)
	 { lu->calc (*jacob);
	   if (condlimit >= 0 && lu->condest(*jacob) > condlimit)
	    { flags |= RX_GENKIN_NEED_SVD;
	    }
	 }
	else if (numj > 6)
	 { JJ->mulT (*jacob, *jacob);
	   lu->calc (*JJ);
	   if (condlimit >= 0 && lu->condest(*JJ) > condlimit*condlimit)
	    { flags |= RX_GENKIN_NEED_SVD;
	    }
	 }
	else // numj < 6
	 { JJ->Tmul (*jacob, *jacob);
	   lu->calc (*JJ);
	   if (condlimit >= 0 && lu->condest(*JJ) > condlimit*condlimit)
	    { flags |= RX_GENKIN_NEED_SVD;
	    }
	 }
	if (flags & RX_GENKIN_NEED_SVD)
	 { if (svd == NULL)
	    { svd = new RxMatdSvd();
	    }
	   svd->calc (*jacob);
	   flags |= RX_GENKIN_JACOBSVD_VALID;
	 }
	flags |= RX_GENKIN_JACOBINV_VALID;
}

RxMatd& RxGenKin::Jset (RxMatd *Jr)
{
	int j;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }
	for (j=0; j<numj; j++)
	 { (*Jr)(0,j) = (*jacob)(0,j);
	   (*Jr)(1,j) = (*jacob)(1,j);
	   (*Jr)(2,j) = (*jacob)(2,j);

	   (*Jr)(3,j) = (*jacob)(3,j);
	   (*Jr)(4,j) = (*jacob)(4,j);
	   (*Jr)(5,j) = (*jacob)(5,j);
	 }
	return (*Jr);
}

RxTwist& RxGenKin::Jmul (RxTwist *tw, const RxVecd &djnts)
{
	double dq;
	int j;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }
	tw->setZero();
	for (j=0; j<numj; j++)
	 { dq = djnts(j);
	   tw->w().x() += (*jacob)(3,j)*dq;
	   tw->w().y() += (*jacob)(4,j)*dq;
	   tw->w().z() += (*jacob)(5,j)*dq;

	   tw->v().x() += (*jacob)(0,j)*dq;
	   tw->v().y() += (*jacob)(1,j)*dq;
	   tw->v().z() += (*jacob)(2,j)*dq;
	 }
	return (*tw);
}

RxVecd& RxGenKin::JTmul (RxVecd *tor, const RxWrench &wr)
{
	int j;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }
	for (j=0; j<numj; j++)
	 { (*tor)(j) = 
		    (*jacob)(0,j)*wr.f().x() +
		    (*jacob)(1,j)*wr.f().y() +
		    (*jacob)(2,j)*wr.f().z() +
		    (*jacob)(3,j)*wr.m().x() +
		    (*jacob)(4,j)*wr.m().y() +
		    (*jacob)(5,j)*wr.m().z();
	 }
	return (*tor);
}

int RxGenKin::invK (RxVecd *, const RxTrsf &, int,
		    const RxVecd *, int)
{
	return (RX_GENKIN_UNIMPLEMENTED);
}

int RxGenKin::Jsolve (RxVecd *dj, const RxTwist &tw)
{
	RxVecd6 twv;
	int i;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }
	if (!(flags & RX_GENKIN_JACOBINV_VALID))
	 { invertJacobian();
	 }
	for (i=0; i<6; i++)
	 { twv(i) = tw(i);
	 }
	if (flags & RX_GENKIN_NEED_SVD)
	 { svd->solve (dj, twv);
	 }
	else
	 { if (numj == 6)
	    { lu->solve (dj, twv);
	    }
	   else if (numj > 6)
	    { RxVecd yv(6);
	      lu->solve (&yv, twv);
	      jacob->Vmul (dj, yv);
	    }
	   else // numj < 6
	    { RxVecd yv(numj);
	      jacob->Vmul (&yv, twv);
	      lu->solve (dj, yv);
	    }
	 }
	return (0);
}

int RxGenKin::JTsolve (RxWrench *wr, const RxVecd &tor)
{
	RxVecd6 wrv;
	int i;

	if (!(flags & RX_GENKIN_JACOBIAN_VALID))
	 { buildJacobian();
	 }
	if (!(flags & RX_GENKIN_JACOBINV_VALID))
	 { invertJacobian();
	 }
	if (flags & RX_GENKIN_NEED_SVD)
	 { svd->Tsolve (&wrv, tor);
	 }
	else
	 { if (numj == 6)
	    { lu->Tsolve (&wrv, tor);
	    }
	   else if (numj > 6)
	    { RxVecd yv(6);
	      jacob->mulV (&yv, tor);
	      lu->solve (&wrv, yv);
	    }
	   else // numj < 6
	    { RxVecd yv(numj);
	      lu->solve (&yv, tor);
	      jacob->mulV (&wrv, yv);
	    }
	 }
	for (i=0; i<6; i++)
	 { (*wr)(i) = wrv(i);
	 }
	return (0);
}
