#include <RcPathKnots.h>
#include <math.h>
#include <mathx.h>
#include <cdefs.h>
#include <stdlib.h>
#include <mutexdefs.h>

#include <RcPathTrajectory.h>

RcPathKnot *RcPathKnot::cache;

DECLARE_STATIC_MUTEX(mux)

#define EXTRA_LIM

// Knot allocation routines. We do our own allocation because it
// allows us to save a reuse knot structures that are already allocated,
// saving all the constructor overhead associated with a call to 'new'.

RcPathKnot *RcPathKnot::allocK()
{ 
	// Allocate a single knot object

	RcPathKnot *kp;
	MUTEX_LOCK(&mux);	// mutex protection for multi-threaded apps
	if ((kp = cache))
	 { cache = kp->link;
	   MUTEX_UNLOCK(&mux);
	   return (kp);
	 }
	MUTEX_UNLOCK(&mux);
	return (new RcPathKnot);
}

RcPathKnot *RcPathKnot::allocKList(int n)
{ 
	// Allocate and return a list of n knot objects, linked
	// together by the 'link' field.

	RcPathKnot *kp, *pp;
	int nsave = n;

	if (n < 1)		// sanity check
	 { return (NULL);
	 }
	MUTEX_LOCK(&mux);	// mutex protection for multi-threaded apps
	pp = kp = cache;
	while (pp && --n)
	 { pp = pp->link;
	 }
	if (pp)
	 { cache = pp->link;
	   pp->link = NULL;
	 }
	else
	 { cache = NULL;
	 }
	MUTEX_UNLOCK(&mux);
	while (n--)
	 { pp = kp;
	   kp = new RcPathKnot;
	   kp->link = pp;
	 }
	return (kp);
}

void RcPathKnot::freeK (RcPathKnot *kp)
{
	// Return a single knot object and put it on the cache list
	// for later reuse.

	MUTEX_LOCK(&mux);	// mutex protection for multi-threaded apps
	kp->link = cache;
	cache = kp;
	MUTEX_UNLOCK(&mux);
}

void RcPathKnot::freeKList (RcPathKnot *kp, RcPathKnot *lp)
{
	// Return a list of knot objects (linked together by the
	// link field) and put them on the cache list for later reuse.

	MUTEX_LOCK(&mux);	// mutex protection for multi-threaded apps
	lp->link = cache;
	cache = kp;
	MUTEX_UNLOCK(&mux);
}

void RcPathKnot::reserve (int nj, int n)
{
	// Preallocates n knots with space for manipulators with nj joints

	RcPathKnot *head, *knot, *tail;

	head = RcPathKnot::allocKList (n);
	for (knot=head; knot; knot=knot->link)
	 { if (knot->link == NULL)
	    { tail = knot;
	    }
	   knot->init (nj);
	 }
	RcPathKnot::freeKList (head, tail);	 
}

char *RcPathKnot::flagStr (char *buf)
{
	// For debugging use. Write info into buf indicating whether
	// certain knot flags are set.

	buf[0] = '\0';
	if (flags & KNOT_INTERPOLATED)
	 { strcat (buf, "interpolated");
	 }
	if (flags & KNOT_CORNER)
	 { strcat (buf, "corner");
	 }
	return (buf);
}

void RcPathKnot::print (const char *str, const char *jtype)
{
	// Used for debugging. Prints out information about a knot.

	RxVecd cxval(numx);
	double jntxAccLim;
	double dels;
	char buf[128];
	int j;

	printf ("%s", str);
	if (idx != -1)
	 { printf (" %3d", idx);
	 }
	printf (" %x next=%x jx=%d s=%9.7f dels=%9.7f %s\n", 
		this, next(), jx, coordx(JS), delcx(JS), flagStr(buf));
	cxval = coordx;
	for (j=0; j<numj; j++)
	 { if (jtype[j] == 'r') cxval(j) *= RTOD;
	 }
	cxval.print ("     coordx   %11.6f\n");
	cxval = delcx;
	for (j=0; j<numj; j++)
	 { if (jtype[j] == 'r') cxval(j) *= RTOD;
	 }
	cxval.print ("     delcx   %11.6f");
	printf (" %11.6f\n", delcx(JS));
	dcxdsR.print ("     dcxdsR %11.6f\n");
	if (next())
	 { next()->dcxdsL.print ("next dcxdsL %11.6f\n");
	 }
	dcxdxR.print ("     dcxdxR %11.6f\n");
	dcxdxL.print ("     dcxdxL %11.6f\n");
	dcxdxA.print ("     dcxdxA %11.6f\n");
	printf ("\n");
}

static double rangeAngRef (double ang, double ref)
{
	// Reduce ang to the range (ref-M_PI, ref+M_PI]

	while (ang <= ref - M_PI)
	 { ang += 2*M_PI;
	 }
	while (ang > ref + M_PI)
	 { ang -= 2*M_PI;
	 }
	return (ang);
}

void rangeJnts (RxVecd *jr, const RxVecd &j1, const RxVecd *ref, 
		const char *jtype, int n)
{
	// jr is set to j1, except that for any joint j which is revolute
	// (as indicated by jtype), jr(j) is set lie within the range
	// (ref(j)-M_PI, ref(j)+M_PI] (if refj is present), or 
	// (-M_PI, M_PI] otherwise.

	int j;

	if (ref)
	 { for (j=0; j<n; j++)
	    { if (jtype[j] == 'r')
	       { (*jr)(j) = rangeAngRef (j1(j), (*ref)(j));
	       }
	      else
	       { (*jr)(j) = j1(j);
	       }
	    }
	 }
	else
	 { for (j=0; j<n; j++)
	    { if (jtype[j] == 'r')
	       { (*jr)(j) = rangeAngRef (j1(j), 0);
	       }
	      else
	       { (*jr)(j) = j1(j);
	       }
	    }
	 }
}

void RcPathKnot::init(int nj)
{
	// Initialize a knot structure for a manipulator with nj 
	// coordinates. This includes allocating storage for the 
	// various coordinate information vectors.

	if (numj != nj)
	 { numx = nj+3;

	   if (numj != 0)
	    { free (store);
	    }
	   numj = nj;

	   store = (double*)malloc(8*sizeof(double)*(numx));

	   coordx.init (numx, store+0);
	   delcx.init (numx, store+numx);
	   dcxdxR.init (numx, store+2*numx);
	   dcxdxL.init (numx, store+3*numx);
	   dcxdxA.init (numx, store+4*numx);
	   d2cxdx2.init (numx, store+5*numx);	
	   dcxdsR.init (numx, store+6*numx);
	   dcxdsL.init (numx, store+7*numx);
	 }
	idx = -1;
	flags = 0;
}

int RcPathKnot::set (RcPathInvKinFxn ifxn, double s, double dels,
		     const char *jtype, int nj, RxVecd *old)
{
	// Set local information for a knot to correspond to the 
	// path position determined by s. Derivatives are determined
	// numerically from values at s+dels.
	
	RxTrsf X2;	// path position at s+del
	RxTrsf Xd;	// difference between X and X2
	double dpos;	// positional difference between X and X2
	double drot;	// rotational difference between X and X2
	int status;	

	init (nj);

	// Use ifxn to determine the X and joint coordinate values
	// at point s on the path.

	status = (*ifxn) (&coordx, &X, old, s);
	if (status < 0)	// bail if necessary
	 { return -1;
	 }
	// Range reduce the revolute joints
	rangeJnts (&coordx, coordx, old, jtype, numj);
	coordx(JS) = s;
	coordx(JP) = 0;	/* will be set later */
	coordx(JR) = 0;	/* will be set later */

	// If X corresponds to a clipped solution, take note of
	// this. At the moment, we don't make use of the INDETERMINATE
	// status value.

	if (status & RC_PATHKIN_BOUNDARY_LIMIT)
	 { flags |= KNOT_CLIPPED;
	 }

	// Compute inverse kinematics at another nearby point s+dels ...

	RxVecd ang2 (numx);
	status = (*ifxn) (&ang2, &X2, &coordx, s+dels);
	if (status < 0)	// bail if necessary
	 { return -1;
	 }
	rangeJnts (&ang2, ang2, &coordx, jtype, numj);

	// ... and use to use this to compute a numerical derivative
	Xd.Imul (X, X2);
	dpos = Xd.p.norm2() * SGN(dels);
	drot = sqrt(Xd.R.o.x()*Xd.R.o.x() + Xd.R.a.x()*Xd.R.a.x() 
				      + Xd.R.a.y()*Xd.R.a.y()) * SGN(dels);
	
	ang2(JS) = s+dels;
	ang2(JP) = dpos;
	ang2(JR) = drot;
	dcxdsR = (ang2 - coordx)/dels;
	dcxdsL = dcxdsR;
	return 1;
}

RcPathKnot::~RcPathKnot()
{
	free (store);
}

void RcPathKnot::setPair (RcPathKnot *next, const RxVecd &range,
			  const char *jtype, int useLen)
{
	// Computes: delcx, jx, dcxdxL, dcxdxR, dcxdxA, and d2cxdx2
	// since all these values depend both on location information
	// at knot, as well as information at the 'next' knot.

	RxTrsf Xd;
	double del;
	double max;
	int j;

	Xd.Imul (X, next->X);
	delcx(JP) = Xd.p.norm2();
	delcx(JR) = sqrt(Xd.R.o.x()*Xd.R.o.x() + Xd.R.a.x()*Xd.R.a.x() 
					  + Xd.R.a.y()*Xd.R.a.y());
	delcx(JS) = (next->coordx(JS)-coordx(JS));

	// Compute delcx, and determine the driving coordinate at
	// the same time.

	jx = (useLen ? JP : JS);
printf ("jx=%d\n", jx);
	max = ABS(delcx(jx)/range(jx));

	for (j=0; j<numj; j++)
	 { del = next->coordx(j) - coordx(j);
	   if (jtype[j] == 'r')
	    { delcx(j) = rangePi(del);
	    }
	   else
	    { delcx(j) = del;
	    }
	   del /= range(j);
	   if (ABS(del) > max)
	    { max = ABS(del);
	      jx = j;
printf ("jx=%d (j)\n", jx);
	    }
	 }

	// now compute derivatives WRT the driving joint

	dcxdxR.scale (1/dcxdsR(jx),dcxdsR);
	dcxdxL.scale (1/next->dcxdsL(jx),next->dcxdsL);
	set2ndDeriv ();
}

void RcPathKnot::set2ndDeriv ()
{
	dcxdxA.scale (1/delcx(jx), delcx);
	d2cxdx2.sub (dcxdxL,dcxdxR);
	d2cxdx2.scale (1/delcx(jx));
}

int RcPathKnot::jntDelOK (RxVecd &maxDel, int defaultJx)
{
	// Tests knot to see if joint differences are within
	// desired bounds.

	int j;

	// only check the joints and the default drive coord

	for (j=0; j<numj; j++)
	 { if (ABS(delcx(j)) > maxDel(j))
	    { checkcode = j;
	      return (0);
	    }
	 }
	if (ABS(delcx(defaultJx)) > maxDel(defaultJx))
	 { checkcode = defaultJx;
	   return (0);
	 }
	return (1);
}

int RcPathKnot::posTolOK (RcPathInvKinFxn ifxn, RcPathFwdKinFxn ffxn,
	double posTol, double rotTol, const char *jtype, RcPathKnot *nextp)
{
	// Tests knot to see if position tolerance is within bounds.

	RxVecd testJnts (numj+3);
	RxTrsf X1;
	RxTrsf X2;
	RxVecd3 raxis;
	double rdist;
	RxTrsf TestX;
	double delx;
	int j;

	// Implements Taylor's algorithm. Compute inverse 
	// kinematics at point half-way between knots

	// Compute position associated with average angle value ...

	delx = delcx(jx)/2;
	for (j=0; j<numj; j++)
	 { testJnts(j) = xToCoord (delx, j);
	 }
	testJnts (JS) = xToCoord (delx, JS);

	(*ffxn) (&X2, testJnts);

	// Now compute the testPnt on the curve

	if (!(flags & KNOT_CLIPPED) && !(nextp->flags & KNOT_CLIPPED))
	 { (*ifxn) ((RxVecd*)0, &TestX, (RxVecd*)0, testJnts(JS));
	 }
	else
	 { RxVecd tempAng (numj);
	   int status;

// XXX XXX XXX Need to worry about indeterminate here

	   status = (*ifxn) (&tempAng, (RxTrsf*)0, &coordx, testJnts(JS));
	   if (status < 0)
	    { return -1;
	    }
	   rangeJnts (&tempAng, tempAng, &coordx, jtype, numj);
	   (*ffxn) (&TestX, tempAng);
	 }
	

	// ... and return true if this is sufficiently close
	// to the true midpoint position 

	X2.R.Imul (TestX.R);					// 20 ..
/*	X2.R.getAxis (&raxis, &rdist); */
	rdist = X2.R.o.x()*X2.R.o.x() + X2.R.a.x()*X2.R.a.x() + X2.R.a.y()*X2.R.a.y();

	X2.p.sub (TestX.p);
	if (X2.p.norm2() > posTol)
	 { checkcode = 0;
	   return (0);
	 }
	else if (rdist > rotTol*rotTol)
	 { checkcode = 1;
	   return (0);
	 }
	else
	 { return (1);
	 }
}

int RcPathKnot::derivLeftOK (const RxVecd &maxAcc, double jntxAccLim,
			     double maxAccErrRatio)
{
	// Tests knot to see if right derivative difference is
	// within bounds.

	double lim;
	int j;

	for (j=0; j<numx; j++)
	 { lim = maxAcc(j)*maxAccErrRatio/(2*jntxAccLim);
	   if (fabs(dcxdxR(j) - dcxdxA(j)) > lim/2)
	    { checkcode = j;
	      return (0);
	    }
	 }
	return (1);
}

int RcPathKnot::derivRightOK (const RxVecd &maxAcc, double jntxAccLim, 
			      double maxAccErrRatio)
{
	// Tests knot to see if left derivative difference is
	// within bounds.

	double lim;
	int j;

	for (j=0; j<numx; j++)
	 { lim = maxAcc(j)*maxAccErrRatio/(2*jntxAccLim);
	   if (fabs(dcxdxL(j) - dcxdxA(j)) > lim/2)
	    { checkcode = j;
	      return (0);
	    }
	 }
	return (1);
}

int RcPathKnot::monoOK ()
{
	// Tests knot to see if s(x) is monotonic

	if (SGN(dcxdxR(JS)) != SGN(delcx(jx)) || 
	    SGN(dcxdxL(JS)) != SGN(delcx(jx)))
	 { return (0);
	 }
	if (3*ABS(dcxdxA(JS)) < ABS(dcxdxR(JS)) + ABS(dcxdxL(JS)) -
			       sqrt(ABS(dcxdxR(JS)*dcxdxL(JS))) )
	 { return (0);
	 }
	return (1);
}

int RcPathKnot::monoLeftOK ()
{
	// left tests to see if s(x) is monotonic

	if (SGN(dcxdxR(JS)) != SGN(delcx(jx)))
	 { return (0);
	 }
	if (ABS(dcxdxR(JS) - dcxdxA(JS)) >= ABS(dcxdxA(JS)))
	 { return (0);
	 }
	return (1);
}

int RcPathKnot::monoRightOK ()
{
	// right tests to see if s(x) is monotonic

	if (SGN(dcxdxL(JS)) != SGN(delcx(jx)))
	 { return (0);
	 }
	if (ABS(dcxdxL(JS) - dcxdxA(JS)) >= ABS(dcxdxA(JS)))
	 { return (0);
	 }
	return (1);
}

double RcPathKnot::estAccBound (const RxVecd &maxVel, 
			   	const RxVecd &maxAcc, int jx)
{
	double max, lim;
	double der;
	int j;

	max = maxAcc(jx);
	for (j=0; j<numx; j++)
	 { if (j != jx)
	    { der = fabs(dcxdxA(j));
	      lim = (maxAcc(j) + maxVel(j)*fabs(d2cxdx2(j))/(der*der))/der;
	      if (lim < max)
	       { max = lim;
	       }
	    }
	 }
	return (max);
}

void RcPathKnot::makeCorner ()
{
	// Make a knot into a corner point. Derivative are recomputed 
	// based on interpolation of adjacent knot data.

	if (prev())
	 { prev()->dcxdxL = 2*prev()->dcxdxA - prev()->dcxdxR;
	   prev()->d2cxdx2 = (prev()->dcxdxL - prev()->dcxdxR)/
			    prev()->delcx(prev()->jx);
	 }
	if (next())
	 { dcxdxR = 2*dcxdxA - dcxdxL;
	   d2cxdx2 = (dcxdxL - dcxdxR)/delcx(jx);
	 }
	flags |= KNOT_CORNER;
}

#define TIME_EPS	(1e-9)

double RcPathKnot::integrate (double *x, RcPathKnot *next, double t)
{
	// Given a knot and a position x of the driving coordinate jx within
	// this knot, integrate x for the time t, or until the end of the
	// knot interval is reached.
	//
	// If the end of the knot interval is reached, return the
	// time remaining and set x to the initial jx value for the 
	// next knot.
	//
	// Otherwise, return the time remaining (which will be negative)
	// and set x to the position reached within the interval.

	int jntx = jx;
	double x0;		// initial value of x
	double v0;		// initial velocity at x0
	double acc;		// acceleration over the interval
	double deltat;		// time from x0 to the end of the interval
	double deltax;		// change in jx over knot interval
	double delta0;		// distance from x0 to end of interval
	double sqrtArg;

	x0 = *x;
	deltax = delcx(jntx);
	delta0 = next->coordx(jntx)-x0;
	if (jx < numj)
	 { delta0 = rangePi(delta0);
	 }
	acc = (next->el - er)/(deltax);
	sqrtArg = 2*next->el - 2*acc*(delta0);
	v0 = SGN(deltax)*sqrt(MAX(0,sqrtArg));
	deltat = 2*(delta0) / (SGN(deltax)*sqrt(MAX(0,2*next->el))+v0);
	if (deltat != deltat)
	 { printf ("num=%g dem=%g x0=%g v0=%g\n",
2*(next->coordx(jntx)-x0), (SGN(deltax)*sqrt(2*next->el)+v0), x0, v0);
	   printf ("deltax=%g 2*next->el=%g\n", deltax, 2*next->el);
	 }
	if (deltat > t)
	 { if (deltat-t < TIME_EPS)
	    { t = deltat;
	      *x = next->coordx(next->jx);
	    }
	   else
	    { *x = x0 + (v0 + acc*t/2)*t;
	    }
	 }
	else
	 { if (next->next())
	    { *x = next->coordx(next->jx);
	    }
	   else
	    { *x = 0;
	    }
	 }
	return (t-deltat);
}

double RcPathKnot::xToCoord (double delx, int j)
{
	/* assume the value of coord(j) WRT x on this knot
	   interval is approximated as a cubic, such that the
	   solution for coord(j) is assumed to have the form

	   a3 y^3 + a2 y^2 + a1 y + a0
	 */

	double dy0, dy1, y0, y1;
	double del;
	double h;

	del = delcx(jx);
	h = delx / del;

	y0 = coordx(j);
	dy0 = dcxdxR(j)*del;
	y1 = next()->coordx(j);
	dy1 = dcxdxL(j)*del;

	return(((-2*(y1-y0)+dy0+dy1)*h + 3*(y1-y0)-2*dy0-dy1)*h + dy0)*h + y0;
}

#if 0
double lamToS (RcPathKnot *knot, double delx, int numj)
{
	/* Same as lamToJval, but customized to return the
	   value associated with s.
	 */

	RcPathKnot *next = knot->next();
	double dy0, dy1, y0, y1;
	double del;
	double h;
	int jx = knot->jx;

	del = knot->delcx(jx);

	h = delx / del;

	y0 = knot->coordx(JS);
	dy0 = del/knot->dcxdsL(jx);
	y1 = next->coordx(JS);
	dy1 = del/next->dcxdsR(jx);

	return(((-2*(y1-y0)+dy0+dy1)*h + 3*(y1-y0)-2*dy0-dy1)*h + dy0)*h + y0;
}
#endif
